/* feature.h
 */
#ifndef _FEATURE_H
#define _FEATURE_H

#include "osl/rating/ratingEnv.h"
#include "osl/state/numEffectState.h"
#include "osl/effect_util/effectUtil.h"
#include "osl/move_classifier/check_.h"
#include "osl/move_classifier/moveAdaptor.h"
#include <string>

namespace osl
{
  namespace rating
  {
    class Feature
    {
      std::string my_name;
    public:
      Feature(const std::string& name) : my_name(name)
      {
      }
      virtual ~Feature();
      virtual bool match(const NumEffectState& state, Move, const RatingEnv&) const =0;
      const std::string& name() const { return my_name; }
    };

    class TakeBack : public Feature
    {
    public:
      TakeBack() : Feature("TakeBack")
      {
      }
      bool match(const NumEffectState&, Move move, const RatingEnv& env) const
      {
	return env.history.hasLastMove() && move.to() == env.history.lastMove().to();
      }
    };

    class TakeBack2 : public Feature
    {
    public:
      TakeBack2() : Feature("TakeBack2")
      {
      }
      bool match(const NumEffectState&, Move move, const RatingEnv& env) const
      {
	return env.history.hasLastMove(2)
	  && move.to() == env.history.lastMove().to()
	  && move.to() == env.history.lastMove(2).to();
      }
    };


    class Check : public Feature
    {
      int property;
    public:
      Check(int p);
      static bool openLong(const NumEffectState& state, Move move) 
      {
	if (move.isDrop())
	  return false;
	return state.hasEffectByPtype<LANCE>(move.player(), move.from())
	  || state.hasEffectByPtype<BISHOP>(move.player(), move.from())
	  || state.hasEffectByPtype<ROOK>(move.player(), move.from());
      }
      bool match(const NumEffectState& state, Move move, const RatingEnv&) const
      {
	using namespace osl::move_classifier;
	if (property == 0 || property == 1) {
	  if (! (PlayerMoveAdaptor<osl::move_classifier::DirectCheck>::isMember(state, move)
		 && ! ConditionAdaptor<osl::move_classifier::OpenCheck>::isMember(state, move)))
	    return false;
	  return property == openLong(state, move);
	}
	if (property == 2)
	  return ! PlayerMoveAdaptor<osl::move_classifier::DirectCheck>::isMember(state, move)
	    && ConditionAdaptor<osl::move_classifier::OpenCheck>::isMember(state, move);
	return PlayerMoveAdaptor<osl::move_classifier::DirectCheck>::isMember(state, move)
	  && ConditionAdaptor<osl::move_classifier::OpenCheck>::isMember(state, move);
      }
      static const CArray<const char*,4> check_property;
    };

    class SendOff : public Feature
    {
    public:
      SendOff() : Feature("SendOff") {}
      bool match(const NumEffectState& state, Move move, const RatingEnv& env) const
      {
	return env.sendoffs.isMember(move.to());
      }
    };

    class Block : public Feature
    {
      int property;
    public:
      Block(int p) : Feature((p == 0 ? "Block" : (p == 1 ? "BlockSelf" : "BlockBoth"))), property(p) {}
      bool match(const NumEffectState& state, Move move, const RatingEnv&) const
      {
	const bool block = state.hasEffectByPtype<LANCE>(alt(move.player()), move.to())
	  || state.hasEffectByPtype<BISHOP>(alt(move.player()), move.to())
	  || state.hasEffectByPtype<ROOK>(alt(move.player()), move.to());
	const bool block_self = state.hasEffectByPtype<LANCE>(move.player(), move.to())
	  || state.hasEffectByPtype<BISHOP>(move.player(), move.to())
	  || state.hasEffectByPtype<ROOK>(move.player(), move.to());
	if (property == 0)
	  return block && !block_self;
	if (property == 1)
	  return !block && block_self;
	return block && block_self;
      }
    };

    // { none, tate, naname, both }
    struct CountOpen
    {
      static int index(const NumEffectState& state, Player player, Position from) 
      {
	if (from.isPieceStand())
	  return -1;
	const bool vertical = state.hasEffectByPtype<LANCE>(player, from)
	  || state.hasEffectByPtype<ROOK>(player, from);
	const bool diagonal = state.hasEffectByPtype<BISHOP>(player, from);
	return diagonal*2+vertical;
      }
    };
    // { none, tate, naname, both } * { none, tate, naname, both }
    class Open : public Feature
    {
      int property;
    public:
      Open(int p) : Feature(name(p)), property(p) {}
      static int index(const NumEffectState& state, Move move)
      {
	if (move.isDrop())
	  return -1;
	return CountOpen::index(state, move.player(), move.from())*4
	  + CountOpen::index(state, alt(move.player()), move.from());
      }
      bool match(const NumEffectState& state, Move move, const RatingEnv&) const
      {
	return index(state, move) == property;
      }
      static const std::string name(int property);
    };

    class ContinueCapture : public Feature
    {
    public:
      ContinueCapture() : Feature("Cont.Capture") {}
      bool match(const NumEffectState& state, Move move, const RatingEnv& env) const
      {
	return env.history.hasLastMove(2) && env.history.lastMove(2).to() == move.from();
      }
    };

    class AttackPin : public Feature
    {
      bool attack;
    public:
      AttackPin(bool a) : Feature(a ? "PinAttack" : "PinDefense"), attack(a) {}
      bool match(const NumEffectState& state, Move move, const RatingEnv& env) const
      {
	if (isMajor(move.ptype()))
	  return false;
	PieceMask pins = (attack ? env.op_pin : env.my_pin);
	while (pins.any()) {
	  const Piece p = state.getPieceOf(pins.takeOneBit());
	  if (state.hasEffectFromTo(move.ptypeO(), move.to(), p.position()))
	    return true;
	}
	return false;
      }
    };

    class Karanari : public Feature
    {
    public:
      Karanari() : Feature("Karanari") {}
      bool match(const NumEffectState& state, Move move, const RatingEnv&) const
      {
	return move.isPromote() && move.capturePtype() == PTYPE_EMPTY
	  && isMajorBasic(move.oldPtype()) && move.from().canPromote(move.player())
	  && move.to().canPromote(move.player());
      }
    };

    class RookDefense : public Feature
    {
    public:
      RookDefense() : Feature("RookDefense") 
      {
      }
      bool match(const NumEffectState& state, Move move, const RatingEnv&) const
      {
	if (move.isDrop())
	  return false;
	Piece rook1 = state.getPieceOf(PtypeTraits<ROOK>::indexMin);
	Piece rook2 = state.getPieceOf(PtypeTraits<ROOK>::indexMin + 1);
	if (move.from() == rook2.position()) 
	  std::swap(rook1, rook2);
	if (move.from() != rook1.position() 
	    || rook2.position().isPieceStand()
	    || rook2.owner() == move.player()
	    || rook2.position().x() != move.to().x())
	  return false;
	return (move.to().y() - rook2.position().y())*playerToMul(move.player()) > 0;
      }
    };

    class BadLance : public Feature
    {
    public:
      BadLance() : Feature("BadLance") 
      {
      }
      bool match(const NumEffectState& state, Move move, const RatingEnv&) const
      {
	if (! (move.isDrop() && move.ptype() == LANCE))
	  return false;
	const Position front = Board_Table.nextPosition(move.player(), move.to(), U);
	return state.hasEffectBy(alt(move.player()), front)
	  && state.hasPieceOnStand(alt(move.player()), PAWN)
	  && !state.isPawnMaskSet(alt(move.player()), front.x());
      }
    };
  }
}


#endif /* _FEATURE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
