/* position.h
 */
#ifndef FEATURE_POSITION_H
#define FEATURE_POSITION_H

#include "osl/rating/feature.h"
#include "osl/move.h"
#include "osl/ptypeTable.h"
#include <string>

namespace osl
{
  namespace rating
  {
    struct DropPtype
    {
      Ptype ptype;
      bool drop;
      DropPtype(Ptype p, bool d) : ptype(p), drop(d)
      {
      }
      bool match(Move m) const
      {
	return (m.ptype() == ptype) && (m.isDrop() == drop);
      }
      static std::string name(Ptype ptype, bool drop) 
      {
	return std::string(Ptype_Table.getCsaName(ptype)) + (drop ? "d" : "m");
      }
      enum { UNIT = PTYPE_MAX+1-PTYPE_PIECE_MIN + (PTYPE_MAX+1 - (PTYPE_BASIC_MIN+1)) };
      static int index(Move move) 
      {
	const Ptype ptype = move.ptype();
	int index;
	if (! isBasic(ptype) || ptype == KING)
	  index = ptype - PTYPE_PIECE_MIN;
	else
	  index = (PTYPE_BASIC_MIN+1-PTYPE_PIECE_MIN) + (ptype - (PTYPE_BASIC_MIN+1))*2 + move.isDrop();
	return index;
      }
    };

    class RelativeKingX : public Feature, DropPtype
    {
      int x;
      bool attack;
    public:
      static const std::string name(int x, bool /*attack*/);
      RelativeKingX(int ix, bool a, Ptype ptype, bool drop) 
	: Feature(name(ix,a)+DropPtype::name(ptype, drop)), DropPtype(ptype, drop), x(ix), attack(a) {}
      static int makeX(bool attack, const NumEffectState& state, Move move) 
      {
	const Position king = state.getKingPosition(attack ? alt(move.player()) : move.player());
	return abs(move.to().x() - king.x());
      }
      bool match(const NumEffectState& state, Move move, const RatingEnv&) const
      {
	if (! DropPtype::match(move))
	  return false;
	return makeX(attack, state, move) == x;
      }
    };

    class RelativeKingY : public Feature, DropPtype
    {
      int y;
      bool attack;
    public:
      static const std::string name(int y, bool /*attack*/);
      RelativeKingY(int iy, bool a, Ptype ptype, bool drop) 
	: Feature(name(iy,a)+DropPtype::name(ptype, drop)), DropPtype(ptype, drop), y(iy), attack(a) {}
      static int makeY(bool attack, const NumEffectState& state, Move move)
      {
	const Position king = state.getKingPosition(attack ? alt(move.player()) : move.player());
	int diff = move.to().y() - king.y();
	if (move.player() == WHITE)
	  diff = -diff;
	return diff;
      }
      bool match(const NumEffectState& state, Move move, const RatingEnv&) const
      {
	if (! DropPtype::match(move))
	  return false;
	return makeY(attack, state, move) == y;
      }
    };

    class PositionX : public Feature, DropPtype
    {
      int x;
      static const std::string name(int x);
    public:
      PositionX(int ix, Ptype ptype, bool drop) 
	: Feature(name(ix)+DropPtype::name(ptype, drop)), DropPtype(ptype, drop), x(ix) {}
      static int makeX(Move move)
      {
	int mx = move.to().x();
	if (mx > 5)
	  mx = 10-mx;
	return mx;
      }
      bool match(const NumEffectState& state, Move move, const RatingEnv&) const
      {
	if (! DropPtype::match(move))
	  return false;
	return x == makeX(move);
      }
    };

    class PositionY : public Feature, DropPtype
    {
      int y;
      static const std::string name(int y);
    public:
      PositionY(int iy, Ptype ptype, bool drop) 
	: Feature(name(iy)+DropPtype::name(ptype, drop)), DropPtype(ptype, drop), y(iy) {}
      static int makeY(Move move) 
      {
	int my = move.to().y();
	if (move.player() == WHITE)
	  my = 10 - my;
	return my;
      }
      bool match(const NumEffectState& state, Move move, const RatingEnv&) const
      {
	if (! DropPtype::match(move))
	  return false;
	return y == makeY(move);
      }
    };
  }
}

#endif /* FEATURE_POSITION_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
