/* group.h
 */
#ifndef _GROUP_H
#define _GROUP_H

#include "osl/rating/feature.h"
#include "osl/rating/range.h"
#include "osl/stl/vector.h"
#include <boost/ptr_container/ptr_vector.hpp>

namespace osl
{
  namespace rating
  {
    /** mutually exclusive set of features */
    struct Group : public boost::ptr_vector<Feature> 
    {
      std::string group_name;
  
      Group(const std::string& name) : group_name(name) {}
      Group(Feature *f) : group_name(f->name()) { push_back(f); }
      virtual ~Group();
      virtual void show(std::ostream&, int name_width, const range_t& range, 
			const vector<double>& weights) const;

      /** @return -1 if not found */
      virtual int findMatch(const NumEffectState& state, Move m, const RatingEnv& env) const;
      void showMinMax(std::ostream& os, int name_width, const range_t& range, 
		      const vector<double>& weights) const;
      void showAll(std::ostream& os, int name_width, const range_t& range, 
		   const vector<double>& weights) const;
      void showTopN(std::ostream& os, int name_width, const range_t& range, 
		    const vector<double>& weights, int n) const;
      void saveResult(const std::string& directory, const range_t& range, 
		      const vector<double>& weights) const;
      bool load(const std::string& directory, const range_t& range, 
		vector<double>& weights) const;
    };

    struct TakeBackGroup : public Group
    {
      TakeBackGroup() : Group("TakeBack")
      {
	push_back(new TakeBack());
	push_back(new TakeBack2());
      }
      void show(std::ostream& os, int name_width, const range_t& range, 
		const vector<double>& weights) const
      {
	showAll(os, name_width, range, weights);
      }
      int findMatch(const NumEffectState& state, Move move, const RatingEnv& env) const
      {
	const Position to = move.to();
	if (! env.history.hasLastMove() || env.history.lastMove().to() != to)
	  return -1;
	if (! env.history.hasLastMove(2) || env.history.lastMove(2).to() != to)
	  return 0;
	return 1;
      }
    };

    struct CheckGroup : public Group
    {
      CheckGroup() : Group("Check")
      {
	for (int i=0; i<4; ++i)
	  push_back(new Check(i));
      }
      void show(std::ostream& os, int name_width, const range_t& range, 
		const vector<double>& weights) const
      {
	showAll(os, name_width, range, weights);
      }
      int findMatch(const NumEffectState& state, Move move, const RatingEnv& env) const
      {
	using namespace osl::move_classifier;
	const bool direct = PlayerMoveAdaptor<osl::move_classifier::DirectCheck>::isMember(state, move);
	const bool open = ConditionAdaptor<osl::move_classifier::OpenCheck>::isMember(state, move);
	int index = -1;
	if (direct && !open)
	  index = Check::openLong(state, move);
	else if (open)
	  index = direct + 2;
	return index;
      }
    };

    struct BlockGroup : public Group
    {
      BlockGroup() : Group("Block")
      {
	push_back(new Block(0));
	push_back(new Block(1));
	push_back(new Block(2));
      }
      void show(std::ostream& os, int name_width, const range_t& range, 
		const vector<double>& weights) const
      {
	showAll(os, name_width, range, weights);
      }
      int findMatch(const NumEffectState& state, Move move, const RatingEnv& env) const
      {
	const bool block = state.hasEffectByPtype<LANCE>(alt(move.player()), move.to())
	  || state.hasEffectByPtype<BISHOP>(alt(move.player()), move.to())
	  || state.hasEffectByPtype<ROOK>(alt(move.player()), move.to());
	const bool block_self = state.hasEffectByPtype<LANCE>(move.player(), move.to())
	  || state.hasEffectByPtype<BISHOP>(move.player(), move.to())
	  || state.hasEffectByPtype<ROOK>(move.player(), move.to());
	const int index = block + block_self*2 -1;
	return index;
      }
    };

    struct OpenGroup : public Group
    {
      OpenGroup() : Group("Open")
      {
	for (int i=0; i<16; ++i)
	  push_back(new Open(i));
      }
      void show(std::ostream& os, int name_width, const range_t& range, 
		const vector<double>& weights) const
      {
	showTopN(os, name_width, range, weights, 3);
      }
      int findMatch(const NumEffectState& state, Move move, const RatingEnv& env) const
      {
	const int index = Open::index(state, move);
	return index;
      }
    };

  }
}

#endif /* _GROUP_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
