#ifndef _PARALLEL_SEARCH_H
#define _PARALLEL_SEARCH_H
#include "osl/misc/lightMutex.h"
#include "osl/misc/atomicCounter.h"
#include "osl/misc/alarm.h"
#include "osl/stl/vector.h"
#include "osl/stl/slist.h"
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <boost/thread.hpp>
#include <boost/cstdint.hpp>
#include <deque>
#include <stdexcept>

#ifndef NOT_ONE_THREAD_PER_PROC
/** ĤΥץå˺륹åɤϤ1Ĥˤ */
#  define ONE_THREAD_PER_PROC
#endif

namespace osl
{
  namespace search
  {
    typedef LightMutex Mutex;

    class Worker;
    class Job;
    struct StopNow : std::runtime_error
    {
      StopNow();
    };
    class JobContent
    {
    public:
      static const int DefaultPriority=100;
      static const int NotFoundPriority=100000;
    private:
      boost::condition condition;
      boost::mutex mutex;
      Worker* worker;
      volatile bool finished;
      int priority;
      uint64_t blocking;
      volatile bool must_stop_now;
      typedef slist<boost::weak_ptr<JobContent> > list_t;
      list_t children;
    public:
      JobContent() : worker(0), finished(0), priority(DefaultPriority),
		     blocking(0),
		     must_stop_now(false)
      {
      }
      explicit JobContent(int priority, uint64_t b=0)
	: worker(0), finished(0), priority(priority), blocking(b),
	  must_stop_now(false)
      {
	assert(priority<NotFoundPriority);
      }
      virtual ~JobContent();
      virtual void operator()(void)=0;
      virtual int id() const { return 0; }
      void setFinished(bool val) {
	boost::mutex::scoped_lock lk(mutex);
	finished=val;
	condition.notify_all();
      }
      void dump() const;
      bool isFinished() const{
	return finished;
      }
      void setWorker(Worker* w) { worker=w; }
      Worker *getWorker() const { return worker; }
      void checkStop();
      void waitResult();
      int getPriority() const{
	return priority;
      }
      uint64_t getBlocking() const{
	return blocking;
      }
      void setPriority(int p) {
	priority=p;
      }
      void stopNow();
      void addChildJob(const boost::shared_ptr<JobContent>& new_child) {
	boost::mutex::scoped_lock lk(mutex);
	children.push_front(new_child);
      }
      bool stopping() const {
	return must_stop_now;
      }
    };
    class ExitNullJobLock
    {
      boost::shared_ptr<JobContent> job_content;
    public:
      ExitNullJobLock(const boost::shared_ptr<JobContent>& job_content)
	: job_content(job_content) {}
      ~ExitNullJobLock() {
	job_content->setWorker(0);
	job_content->setFinished(true);
      }
    };
    class Job
    {
      boost::shared_ptr<JobContent> job_content;
    public:
      Job() {}
      explicit Job(const boost::shared_ptr<JobContent>& job_content) 
	: job_content(job_content) {
      }
      bool isNull() const{
	return !job_content;
      }
      JobContent* getContent() { return job_content.get(); }
      void operator()(void) {
	ExitNullJobLock lock(job_content);
	assert(!isNull());
	try{
	  (*job_content)();
	}
	catch (StopNow& e) {
	}
      }
      void setWorker(Worker* w) {
	job_content->setWorker(w);
      }
      void setFinished(bool val) {
	job_content->setFinished(val);
      }
      void waitResult() {
	job_content->waitResult();
      }
      int getPriority() const{
	return job_content->getPriority();
      }
      uint64_t getBlocking() const{
	return job_content->getBlocking();
      }
      void stopNow() const {
	job_content->stopNow();
      }
      void checkStop() const {
	job_content->checkStop();
      }
      void addChildJob(const boost::shared_ptr<JobContent>& new_child) {
	if (job_content)
	  job_content->addChildJob(new_child);
      }
      friend class JobPool;
    };
    class JobPool
    {
      std::list<Job> jobs;
      friend class ParallelSearch;
    public:
      typedef std::list<Job>::iterator iterator;
      typedef std::list<Job>::const_iterator const_iterator;
      void push_back(Job job) {
	jobs.push_back(job);
      }
      bool pop_back() {
	if (jobs.empty())
	  return false;
	jobs.pop_back();
	return true;
      }
      bool pop_back(JobContent *content) {
	for (iterator p=jobs.begin(); p!=jobs.end(); ++p) {
	  if (p->job_content.get() == content) {
	    jobs.erase(p);
	    return true;
	  }
	}
	return false;
      }
      int getPriority() const{
	if (jobs.empty()) 
	  return JobContent::NotFoundPriority;
	return jobs.front().getPriority();
      }
      int getPriorityMax() const{
	if (jobs.empty()) 
	  return -1;
	return jobs.back().getPriority();
      }
      /** call with shared_mutex locked */
      const Job getJob(Worker *worker) {
	if (jobs.empty()) 
	  return Job();
	Job ret=jobs.front();
	jobs.pop_front();
	ret.setWorker(worker);
	return ret;
      }
      // iterator ֤ shared_mutex lock
      iterator getBlockingJob(uint64_t, uint64_t);
      const Job getJob(Worker *worker, iterator p) 
      {
	assert(p!=jobs.end());
	Job ret = *p;
	jobs.erase(p);
	ret.setWorker(worker);
	return ret;
      }
      iterator end() { return jobs.end(); }
      void clear(){
	jobs.clear();
      }
    };
    class ParallelSearch;
    /**
     * 
     */
    class Worker
    {
      ParallelSearch* parent;
      boost::thread *th;
      boost::shared_ptr<JobPool> job_pool;
      Job current_job;
      unsigned int thread_id;
      volatile bool waiting;
      volatile uint64_t waiting_block;
      volatile uint64_t sibling;
    public:
      explicit Worker(ParallelSearch* parent,boost::thread* th=0) 
	: parent(parent),th(th),job_pool(new JobPool()),
	  thread_id(static_cast<unsigned int>(-1)), waiting(true), waiting_block(0)
      {
      }
      void operator()(void);
      /** call with shared_mutex locked */
      const Job getJob(Worker *w) {
	Job ret=job_pool->getJob(w);
	return ret;
      }
      int getPriority() const{
	return job_pool->getPriority();
      }
      int getPriorityMax() const{
	return job_pool->getPriorityMax();
      }
      boost::thread* getThread() const{
	return th;
      }
      /** shared_mutex  lock  */
      void push_back(Job job);
      /** shared_mutex  lock  */
      bool pop_back_job();
      bool pop_back_job(JobContent*);
    public:
      boost::mutex block_mutex;
      boost::condition block_condition;
      
      void checkStop() const;
      const Job getCurrentJob() const{
	return current_job;
      }
      void setCurrentJob(Job const& job) {
	current_job=job;
      }
      void resetCrurrentJob() {
	current_job=Job();
      }
      unsigned int threadId() const { return thread_id; }
      friend class ParallelSearch;

      void stopMaster();
      bool isWaiting() const { return waiting; }
      void setWaiting(bool wait) { waiting = wait; }
      void setWaitingBlock(uint64_t block, uint64_t sib) { waiting = true; waiting_block = block; sibling = sib; }
      void resetWaitingBlock() { waiting = false; waiting_block = 0; }
      uint64_t waitingBlock() const { return waiting_block; }
      uint64_t waitingSibling() const { return sibling; }
    private:
      void setThreadId(unsigned int id) { thread_id = id; }
      /** shared_mutex  lock ֤ǸƤ */
      void clearJobs() { job_pool->clear(); };
    };

    class SetCurrentJobLock
    {
      Worker& worker;
      Job backup;
    public:
      SetCurrentJobLock(Worker& w,Job const& job) :worker(w),backup(w.getCurrentJob()){
	worker.setCurrentJob(job);
      }
      ~SetCurrentJobLock(){
	worker.setCurrentJob(backup);
      }
    };

    struct DecActiveHelper{
      ParallelSearch & parallel_search;
      explicit DecActiveHelper(ParallelSearch & parallel_search);
      ~DecActiveHelper();
    };
    struct IncActiveHelper{
      ParallelSearch & parallel_search;
      explicit IncActiveHelper(ParallelSearch & parallel_search);
      ~IncActiveHelper();
    };
    struct IncWaitingHelper{
      ParallelSearch & parallel_search;
      explicit IncWaitingHelper(ParallelSearch & parallel_search);
      ~IncWaitingHelper();
    };
    struct DecWaitingHelper{
      ParallelSearch & parallel_search;
      explicit DecWaitingHelper(ParallelSearch & parallel_search);
      ~DecWaitingHelper();
    };

    class ParallelSearch
    {
      AtomicCounter num_actives;
      AtomicCounter max_num_actives;
      AtomicCounter thread_count;
      AtomicCounter waiting; // WorkerΥȥåץ٥ԤäƤ륹åɿ
      vector<Worker*> workers;
      int num_cpus;
      time_t last_time;
      AlarmSwitch master_stop_flag;
      bool verbose;
      volatile bool stop_all;
      void checkTime();
    public:
      static const int MaxCPUs = 16;

      boost::condition shared_condition;
      boost::mutex shared_mutex;

      ParallelSearch();
      ~ParallelSearch();

      void setVerbose(bool v=true) { verbose = v; }
      void addInLock(Worker* worker) {
	workers.push_back(worker);
	worker->setThreadId(workers.size()-1);
      }
      void add(Worker* worker) {
	boost::mutex::scoped_lock lk(shared_mutex);
	addInLock(worker);
      }
      const Job getJob(Worker *w);
      const Job getJobInLock(Worker *w);
      const Job getJobNonBlock(Worker *w);
      /** blocking bitΩäƤСbitjobΤ */
      const Job getJobNonBlock(Worker *w, uint64_t blocking, uint64_t sibling);
      /**
       * ߤΥåɤWorker֤
       * бΤʤϡallocate롥
       */
      Worker* getWorker();
      void notify_one() {
	shared_condition.notify_one();
      }
      void makeWorkers();
      void setNumCPUs(int n) {
	assert(n <= MaxCPUs);
	num_cpus=n;
      }
      int numCpus() const { return num_cpus; }
      void checkStop();
      bool stopping() const { 
	return stop_all 
	  || (master_stop_flag.notify
	      && *master_stop_flag.notify == Alarm::TIMEOUT);
      }
      void waitAll();

      void setMasterAlarm(AlarmSwitch alarm) { master_stop_flag = alarm; }
      bool stopMaster();
      void reportWorkers();

      int numActives() const { return num_actives.value(); }
      /* root waitƤworkertrueꡢwho == 0.
       * block Ƥͤ who ˥ɥ쥹 
       */
      bool hasWaiting(uint64_t blocking, Worker *& who);
    private:
      friend class JobContent;
      friend class IncActiveHelper;
      friend class DecActiveHelper;
      friend class IncWaitingHelper;
      friend class DecWaitingHelper;
      void incActives();
      void decActives();
      void incWaiting();
      void decWaiting();
    };

    extern osl::search::ParallelSearch parallelSearch;
  } // namespace search
} // namespace osl
#endif /* _PARALLEL_SEARCH_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
