/* searchState.h
 */
#ifndef SEARCHSTATE_H
#define SEARCHSTATE_H

#include "osl/search/killerMoveTable.h"
#include "osl/search/bigramKillerMove.h"
#include "osl/search/historyTable.h"
#include "osl/search/recordStack.h"
#include "osl/search/firstMoveThreatmate.h"
#include "osl/checkmate/dualCheckmateSearcher.h"
#include "osl/checkmate/fixedDepthSearcher.h"
#include "osl/effect_util/neighboring8Direct.h"
#include "osl/state/numEffectState.h"
#include "osl/hash/hashKey.h"
#include "osl/repetitionCounter.h"
#include "osl/container/moveStack.h"
#include "osl/apply_move/applyMoveWithPath.h"
#include "osl/misc/alarm.h"

#ifdef OSL_SMP
# include "osl/search/parallelSearch.h"
#endif

#include <boost/shared_ptr.hpp>
#include <boost/utility.hpp>

namespace osl
{
  namespace search
  {
    class Worker;
    /**
     * õ򤹤˶ͭ
     */
    struct SearchStateShared : boost::noncopyable
    {
      BigramKillerMove bigram_killers;
      HistoryTable history_table;

      SearchStateShared();
      ~SearchStateShared();
    };
    /**
     * QuiescenceSearch Ǥɬפʬ
     */
    class SearchStateCore
    {
    public:
      typedef DualCheckmateSearcher<> checkmate_t;
    private:
      NumEffectState current_state;
      checkmate_t* checkmate_searcher;
      const Worker *worker_cache;
#ifdef OSL_SMP
    public:
      void setCheckmateSearcher(checkmate_t *new_checkmate,
				const Worker *new_worker) {
	checkmate_searcher = new_checkmate;
	worker_cache = new_worker;
      }
    private:
#endif
      PathEncoding current_path;
      MoveStack move_history;
      KillerMoveTable killer_moves;
      int root_depth;
    protected:
      RecordStack record_stack;
      RepetitionCounter repetition_counter;
      boost::shared_ptr<SearchStateShared> shared;
      /**
       * ͤˤäƤõ
       */
      volatile int stop_flag;
    public:
      SearchStateCore(const NumEffectState& s, checkmate_t& checker);
      virtual ~SearchStateCore();
      int curDepth() const { return history().size() - root_depth; }

      // timer ط
      const AlarmSwitch alarmSwitch() { return AlarmSwitch(&stop_flag); }
      void setTimeOut(int seconds);
      void resetTimeOut();
      bool timerAvailable() const { return stop_flag == Alarm::WATCHING; }
    private:
      void throwTimeOut();
    public:
      void testTimeOut() { 
	if (stop_flag == Alarm::TIMEOUT)
	  throwTimeOut();
#ifdef OSL_SMP
	if (worker_cache)
	  worker_cache->checkStop();
#endif
      }
      /**
       * state ΥԡԤ.
       *
       * this->state õλ¸뤬õ exception 
       * ˲Ƥ
       */
      virtual void setState(const NumEffectState& s);
      void setHistory(const MoveStack& h);
      bool hasLastRecord(unsigned int n=0) const
      {
	return record_stack.hasLastRecord(n);
      }
      SimpleHashRecord** lastRecordPtr(unsigned int n=0) 
      {
	return record_stack.lastRecordPtr(n); 
      }
      SimpleHashRecord* lastRecord(unsigned int n=0) 
      {
	return *lastRecordPtr(n); 
      }
      SimpleHashRecord *rootRecord()
      {
	assert(! record_stack.empty());
	return *record_stack.lastRecordPtr(record_stack.size()-1);
      }
      void setRootRecord(SimpleHashRecord *root)
      {
	assert(! record_stack.empty());
	*record_stack.lastRecordPtr(record_stack.size()-1) = root; 
      }

      // killer move
      void setKillerMove(Move best_move)
      {
	if (best_move.isPass())
	  return;
	killer_moves.setMove(curDepth(), best_move);
	const Move last_move = lastMove();
	if (! last_move.isInvalid())
	{
	  shared->bigram_killers.setMove(last_move, best_move);
	}
      }
      void getBigramKillerMoves(MoveVector& moves) const
      {
	shared->bigram_killers.getMove(state(), lastMove(), moves);
#ifdef TRI_PLY_BIGRAM_KILLERS
	if (move_history.hasLastMove(3))
	  shared->bigram_killers.getMove(state(), lastMove(3), moves);
#endif
      }
      void getKillerMoves(MoveVector& moves) const
      {
	getBigramKillerMoves(moves);	
	killer_moves.getMove(state(), curDepth(), moves);
      }
      const BigramKillerMove& bigramKillerMove() const { 
	return shared->bigram_killers;
      }
      void setBigramKillerMove(const BigramKillerMove& killers);
      const HistoryTable& historyTable() const
      {
	return shared->history_table;
      }
      
      // doUndo
      void pushPass()
      {
	const Move pass = Move::PASS(current_state.getTurn());
	current_state.changeTurn();
	current_path.pushMove(pass);
	move_history.push(pass);
      }
      void popPass()
      {
	const Move pass = Move::PASS(alt(current_state.getTurn()));
	current_state.changeTurn();
	current_path.popMove(pass);
	move_history.pop();
      }
    private:
      /**
       * ApplyMove˹Ԥ
       */
      void pushBeforeApply(Move move, SimpleHashRecord**record)
      {
	move_history.push(move);
	record_stack.push(record);
      }
      struct Updator
      {
	const HashKey& new_hash;
	SearchStateCore *state;
	Updator(const HashKey& h, SearchStateCore *s)
	  : new_hash(h), state(s)
	{
	}
	void update()
	{
	  state->updateRepetitionCounterAfterMove(new_hash);
	}
      };
      template <class Function, class Eval>
      struct UpdateWrapper : public Updator
      {
	Function& function;
	Eval& eval;
	Move last_move;
	UpdateWrapper(const HashKey& h, SearchStateCore *s, Function& f, Eval& e, Move l)
	  : Updator(h, s), function(f), eval(e), last_move(l)
	{
	}
	void operator()(Position to)
	{
	  update();
	  eval.update(state->state(), last_move);
	  function(to);
	}
      };
      friend class Updator;
      /**
       * pushBeforeApply θ塤ApplyMove桤FunctionƤ˸ƤФ
       */
      void updateRepetitionCounterAfterMove(const HashKey& new_hash)
      {
	repetition_counter.push(new_hash, current_state);
      }
      /**
       * ApplyMoveθ˹Ԥ
       */
      void popAfterApply()
      {
	record_stack.pop();
	repetition_counter.pop();
	move_history.pop();
      }
    public:
      /**
       * ޤȤdoUndo
       */
      template <Player P, class Function, class Eval>
      void doUndoMoveOrPass(const HashKey& new_hash,
			    Move move, SimpleHashRecord **move_record, 
			    Function& f, Eval& e)
      {
	pushBeforeApply(move, move_record);
	UpdateWrapper<Function, Eval> wrapper(new_hash, this, f, e, move);
	ApplyMoveWithPath<P>::doUndoMoveOrPass(current_state, current_path, 
					       move, wrapper);
	popAfterApply();
      }
      const Move lastMove(int i=1) const { return move_history.lastMove(i); }
      const MoveStack& history() const { return move_history; }
      const RecordStack& recordHistory() const { return record_stack; }
      const PathEncoding& path() const { return current_path; }
      const NumEffectState& state() const { return current_state; }
      const checkmate_t& checkmateSearcher() const { return *checkmate_searcher; }
      const RepetitionCounter& repetitionCounter() const { 
	return repetition_counter;
      }
      const HashKey& currentHash() const
      {
	return repetition_counter.history().top();
      }

      /**
       * ̲ doUndo , hash 򹹿ʤ
       */
      template <Player P, class Function>
      void doUndoMoveLight(Move move, Function& f)
      {
	ApplyMoveWithPath<P>::doUndoMoveOrPass
	  (static_cast<NumEffectState&>(current_state), current_path, 
	   move, f);
      }

      template <Player P>
      bool isLosingState(int node_limit)
      {
	assert(P == path().turn());
	const bool lose = 
	  checkmate_searcher->isLosingState<P>
	  (node_limit, current_state, currentHash(), path(), lastMove());
	return lose;
      }
    public:
      template <Player P>
      bool isWinningState(int node_limit, Move& checkmate_move, 
			  AttackOracleAges& age)
      {
	assert(P == path().turn());
	const bool win = checkmate_searcher->isWinningState<P>
	  (node_limit, current_state, currentHash(), path(), checkmate_move, age,
	   lastMove());
	return win;
      }
      /** FixedDepthSearcher Ƥ */
      template <Player P>
      bool isWinningStateShort(int depth, Move& checkmate_move)
      {
	checkmate::FixedDepthSearcher searcher(current_state);
	const ProofDisproof pdp=searcher.hasCheckmateMove<P>(depth, checkmate_move);
	return pdp.isCheckmateSuccess();
      }
      /**
       * P μ֤Pζ̤˵ͤäƤ뤫ɤ
       */
      template <Player P>
      bool isThreatmateState(int node_limit, Move& threatmate_move, 
			     AttackOracleAges& age)
      {
	assert(P == path().turn());
	current_state.changeTurn();
	HashKey threatmate_hash = currentHash();
	threatmate_hash.changeTurn();
	const PathEncoding threatmate_path(current_state.getTurn());
	const Player Opponent = PlayerTraits<P>::opponent;
	const bool threatmate 
	  = checkmate_searcher->template isWinningState<Opponent>
	  (node_limit, current_state, 
	   threatmate_hash, threatmate_path, threatmate_move, age);
	current_state.changeTurn();
	return threatmate;
      }
      template <Player P>
      bool isThreatmateStateShort(int depth, Move& threatmate_move)
      {
	assert(P == path().turn());
	current_state.changeTurn();

	const Player Opponent = PlayerTraits<P>::opponent;

	checkmate::FixedDepthSearcher searcher(current_state);
	const ProofDisproof pdp
	  = searcher.hasCheckmateMove<Opponent>(depth, threatmate_move);

	current_state.changeTurn();
	return pdp.isCheckmateSuccess();
      }
      virtual bool abort(Move) const;

      bool tryThreatmate() const 
      {
	const Move last_move = lastMove();
	if (! last_move.isNormal())
	  return false;
	const Position king = state().getKingPosition(state().getTurn());
	if (curDepth() == 1)
	  return FirstMoveThreatmate::isMember(last_move, king);
	return Neighboring8Direct::hasEffect
	  (state(), last_move.ptypeO(), last_move.to(), king);

      }

    };
    
    /**
     * SearchFramework Τtemplate parameter ޤޤʤʬ.
     */
    class SearchState : public SearchStateCore
    {
    public:
      /** õ䡤ؼǤΨι⤤꤬äȤ̵뤹ϰ */
      static const int ReSearchLimitMargin = 80;
    protected:
      int root_limit;
      int cur_limit;
    public:
      SearchState(const NumEffectState& s, checkmate_t& checker);
      virtual ~SearchState();

      void setState(const NumEffectState& s);
      void setKillerMove(Move best_move)
      {
	if (best_move.isPass())
	  return;
	shared->history_table.setMove(curLimit(), best_move);
	SearchStateCore::setKillerMove(best_move);
      }
      
      int curLimit() const { return cur_limit; }

      bool abort(Move) const;

    protected:
      /**
       * root  limitͤõϤ뤳Ȥ
       */
      void setRoot(int limit)
      {
	root_limit = limit;
	cur_limit = limit;
      }
      void addLimit(int limit) { cur_limit += limit; assert(cur_limit >= 0); }
      void subLimit(int limit) { cur_limit -= limit; assert(cur_limit >= 0); }

      /** μΤƶϢ³2롼פޤǿ 
       * @param history_max ʾվʤ
       */
      int countSacrificeCheck2(int history_max) const;
      /** debug  */
      void checkPointSearchAllMoves();
    };
  } // namespace search
} // namespace osl

#endif /* SEARCHSTATE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
