/* bigramPosition.cc
 */
#include "osl/apply_move/applyMove.h"
#include "osl/category/bigramPosition.h"
#include "osl/container/moveLogProbVector.h"
#include "osl/effect_util/neighboring8Direct.h"
#include "osl/oslConfig.h"
#include <iostream>

namespace osl
{
  namespace category
  {
    BigramTable BigramPosition::probTable(OslConfig::home_c_str(), "prev");
    BigramTable SamePlayerBigramPosition::probTable(OslConfig::home_c_str(), "prev-prev");
    KingNeighborBigramTable BigramAttack::probTable(OslConfig::home_c_str(), "attack");
    KingNeighborBigramTable BigramDefense::probTable(OslConfig::home_c_str(), "defense");
    BigramHistoryTable BigramHistory::probTable(1);
    static BigramHelper bigramHelper = BigramHelper();
  }
}

osl::category::BigramHelper::
BigramHelper()
{
  for (size_t i = 0; i < osl::PieceStand::order.size(); i++)
  {
    pieceMap[osl::PieceStand::order[i]] = i;
  }
}

int osl::category::BigramHelper::
moveToIndex(osl::Move move, bool from) const
{
  osl::Position p = from ? move.from() : move.to();

  if (p.isPieceStand())
    return 100 + pieceMap[osl::unpromote(move.ptype())];
  else
    return (p.x() * 10) + p.y();
}


int osl::category::BigramHelper::
moveToIndex(osl::Move move, osl::Player player) const
{
  assert(move.isNormal());
  int value = (moveToIndex(move, true) << 8
	       | (moveToIndex(move, false)));
  if (player == osl::BLACK)
    value |= 1 << 31;

  return value;
}

int osl::category::BigramHelper::
moveToValue(osl::Move move) const
{
  assert(move.isNormal());
  int value = (moveToIndex(move, true) << 8
	       | (moveToIndex(move, false)));
  return value;
}

int osl::category::BigramHelper::
toIndex(osl::Position pos, osl::Move move, osl::Player player) const
{
  assert(move.isNormal());
  int value = ((pos.x() * 10 + pos.y()) << 16 
	       | moveToIndex(move, true) << 8
	       | (moveToIndex(move, false)));
  if (player == osl::BLACK)
    value |= 1 << 31;

  return value;
}

osl::Move osl::category::BigramHelper::
indexToMove(int value, const CategoryEnv::effect_state_t& state) const
{
  int from, to, mask;
  mask = (1 << 8) - 1;
  from = (value >> 8) & mask;
  to = value & mask;
  Position toP(to / 10, to % 10);
  if (from >= 100)
  {
    if (!state.getPieceAt(toP).isEmpty())
      return Move::INVALID();
    Ptype type = PieceStand::order[from - 100];
    Move move(toP, type, state.getTurn());
    if (! Ptype_Table.canDropTo(state.getTurn(), type, toP))
      return Move::INVALID();
    if (state.isAlmostValidMove<false>(move))
      return move;
    else
      return Move::INVALID();
  }
  else
  {
    Position fromP(from / 10, from % 10);
    Piece p = state.getPieceAt(fromP);
    Piece toPiece = state.getPieceAt(toP);

    if (p.isEmpty() || p.owner() != state.getTurn()
	|| (!toPiece.isEmpty() && toPiece.owner() == state.getTurn()))
      return Move::INVALID();

    if (! state.hasEffectByPiece(p, toP))
      return Move::INVALID();

    Move move = Move(fromP, toP, p.ptype(),
		     toPiece.ptype(), false, state.getTurn());
    if (!isPromoted(p.ptype())
	&& ! Ptype_Table.canDropTo(state.getTurn(), p.ptype(), toP))
    {
      move = Move(fromP, toP, promote(p.ptype()),
		  toPiece.ptype(), true, state.getTurn());
    }

    if (state.isAlmostValidMove<false>(move))
      return move;
    else
      return Move::INVALID();
  }
}


osl::category::BigramTable::
BigramTable() : table(new map_t)
{
}

osl::category::BigramTable::
BigramTable(const char *oslHome, const char *name) : table(new map_t)
{
  std::string path(oslHome);

  path += "/data/bigram/";
  path += name;
  path += ".txt";
  init(path.c_str());
}

void osl::category::BigramTable::
init(const char *filename)
{
  min = osl::category::CategoryTraits::maxProb();
  FILE *fp = fopen(filename, "r");
  if (!fp)
  {
    std::cerr << "Failed to open bigram table " << filename << std::endl;
    return;
  }

  while (true)
  {
    CArray<char,2> turn;
    int from1, to1, from2, to2, value;
    int ret;
    ret = fscanf(fp, "%1s %d %d %d %d %d\n", &turn[0],
		 &from1, &to1, &from2, &to2, &value);
    if (ret == EOF)
      break;
    else if (ret < 6)
    {
      std::cerr << "Unknown Entry " << from1 << " " << to1 << std::endl;
    }
    min = std::min(min, value);

    int key = (from1 << 8) | to1;
    int pos = (from2 << 8) | to2;
    if (turn[0] == '+')
    {
      key |= 1 << 31;
    }

    (*table)[key].push_back(std::make_pair(pos, value));
  }
  if (fp)
    fclose(fp);
}

void osl::category::BigramTable::
generate(const CategoryEnv& env, MoveLogProbVector& out, int back) const
{
  assert(back == 1 || back == 2);
  const osl::Move key_move = env.history->lastMove(back);
  if (!key_move.isNormal())
    return;
  int index = bigramHelper.moveToIndex(key_move, env.state->getTurn());
  if (table->find(index) != table->end())
  {
    typedef osl::stl::vector<std::pair<int, int> > vector_t;
    const vector_t& moves = table->find(index)->second;
    for (size_t i = 0; i < moves.size(); i++)
    {
      if (moves[i].second > env.limit)
	break;
      
      const Move move = bigramHelper.indexToMove(moves[i].first, *env.state);
      if (move.isInvalid())
	continue;
      if (!move.from().isPieceStand()
	  && move.ptype() != KING
	  && move.ptype() != GOLD
	  && !isPromoted(move.ptype())
	  && (move.from().canPromote(move.player())
	      || move.to().canPromote(move.player())))
      {
	const Move promoted_move(move.from(), move.to(),
				 promote(move.ptype()),
				 move.capturePtype(), true,
				 move.player());
	assert(env.state->isAlmostValidMove(promoted_move));
	out.push_back(promoted_move,
		      moves[i].second);
      }
      out.push_back(move, moves[i].second);
    }
  }
}


osl::category::BigramHistoryTable::BigramHistoryTable(int) : table(new map_t)
{
}

void osl::category::BigramHistoryTable::
setupTable(const SimpleState& initialState,
	   const CategoryEnv::effect_state_t& eState,
	   const container::MoveStack& history)
{
  table->clear();

  CategoryEnv::effect_state_t state(initialState);
  container::MoveStack genHistory;

  for (size_t i = 0; ; i++)
  {
    const int decay = DECAY_VALUE * (history.size() - i) / 2;
    if (eState.getTurn() == state.getTurn()
	&& RETAIN_REALIZATION_PROBABILITY > decay)
    {
      CategoryEnv env(&state, 200, &genHistory);
      container::MoveLogProbVector moves;
      BigramPosition::generate(env, moves);
      SamePlayerBigramPosition::generate(env, moves);

      for (size_t i = 0; i < moves.size(); i++)
      {
	assert(!moves[i].getMove().isInvalid());
	int index = bigramHelper.moveToIndex(moves[i].getMove(),
					     eState.getTurn());
	if (moves[i].getLogProb() + decay < RETAIN_REALIZATION_PROBABILITY
	    && (table->find(index) == table->end()
		|| (*table)[index] > moves[i].getLogProb() + decay)
	    && !bigramHelper.indexToMove(index, eState).isInvalid())
	{
	  (*table)[index] = moves[i].getLogProb() + decay;
	}
      }
    }
    // 3ܰʹѤ̾ bigram ȽŤʤ뤬ߤζ̤μϿ
    if (i == history.size())
      break;
    const Move move = history.lastMove(history.size() - i);
    apply_move::ApplyMoveOfTurn::doMove(state, move);
    genHistory.push(move);
  }
}

void osl::category::BigramHistoryTable::
generate(const CategoryEnv& env, MoveLogProbVector& out) const
{
  if (env.depth > 3 || env.depth == 0)
    return;

  for (map_t::iterator it = table->begin();
       it != table->end();
       ++it)
  {
    const Move move = bigramHelper.indexToMove(it->first, *env.state);
    if (move.isInvalid())
    {
      continue;
    }
    if (it->second <= env.limit)
    {
      assert(env.state->isAlmostValidMove(move));
      out.push_back(move, it->second);
      if (!move.from().isPieceStand()
	  && move.ptype() != KING
	  && move.ptype() != GOLD
	  && !isPromoted(move.ptype())
	  && (move.from().canPromote(move.player())
	      || move.to().canPromote(move.player())))
      {
	Move promoted_move(move.from(), move.to(), promote(move.ptype()),
			   move.capturePtype(), true, move.player());
	assert(env.state->isAlmostValidMove(promoted_move));
	out.push_back(promoted_move, it->second);
      }
    }
  }
}

osl::category::KingNeighborBigramTable::
KingNeighborBigramTable() : table(new map_t)
{
}

osl::category::KingNeighborBigramTable::
KingNeighborBigramTable(const char *oslHome, const char *name) : table(new map_t)
{
  std::string path(oslHome);

  path += "/data/bigram/";
  path += name;
  path += ".txt";
  init(path.c_str());
}

void osl::category::KingNeighborBigramTable::
init(const char *filename)
{
  min = osl::category::CategoryTraits::maxProb();
  FILE *fp = fopen(filename, "r");
  if (!fp)
  {
    std::cerr << "Failed to open bigram table " << filename << std::endl;
    return;
  }

  while (true)
  {
    CArray<char,2> turn;
    int king, from1, to1, from2, to2, value;
    int ret;
    ret = fscanf(fp, "%1s %d %d %d %d %d %d\n", &turn[0],
		 &king, &from1, &to1, &from2, &to2, &value);
    if (ret == EOF)
      break;
    else if (ret < 7)
    {
      std::cerr << "Unknown Entry " << from1 << " " << to1 << std::endl;
    }
    min = std::min(min, value);

    int key = (king << 16) | (from1 << 8) | to1;
    int pos = (from2 << 8) | to2;
    if (turn[0] == '+')
    {
      key |= 1 << 31;
    }

    (*table)[key].push_back(std::make_pair(pos, value));
  }
  if (fp)
    fclose(fp);
}

bool osl::category::KingNeighborBigramTable::
isIn5x5(const osl::Position kingPosition,
	const osl::Position targetPosition) const
{
  int x = kingPosition.x();
  int y = kingPosition.y();

  // X 
  if (x < 3)
    x = 3;
  if (x > 7)
    x = 7;

  return ((x - 2 <= targetPosition.x() && targetPosition.x() <= x + 2)
	  && (y - 2 <= targetPosition.y() && targetPosition.y() <= y + 2));
}


bool osl::category::KingNeighborBigramTable::
isAttackByMajorPiece(const osl::category::CategoryEnv::effect_state_t& state,
		     const osl::Position pos,
		     const osl::Ptype ptype,
		     const osl::Player turn) const
{
  return ((unpromote(ptype) == osl::ROOK
	   && pos.canPromote(alt(turn)))
	  || (unpromote(ptype) == osl::BISHOP
	      && osl::effect_util::Neighboring8Direct::hasEffect(state,
								 osl::newPtypeO(state.getTurn(), ptype),
								 pos, state.getKingPosition(turn))));
}

void osl::category::KingNeighborBigramTable::
generate(const CategoryEnv& env, MoveLogProbVector& out,
	 int back) const
{
  assert(back == 1 || back == 2);
  const osl::Move key_move = env.history->lastMove(back);

  if (!key_move.isNormal())
    return;
  const osl::Player kingTurn = back == 2 ? alt(env.state->getTurn()) : env.state->getTurn();
  const osl::Position kingPosition = env.state->getKingPosition(kingTurn);

  if (!isIn5x5(kingPosition, key_move.to())
      && !isAttackByMajorPiece(*env.state, key_move.to(), key_move.ptype(),
			       kingTurn))
    return;
  int index = bigramHelper.toIndex(kingPosition, key_move, env.state->getTurn());
  if (table->find(index) != table->end())
  {
    typedef osl::stl::vector<std::pair<int, int> > vector_t;
    const vector_t& moves = table->find(index)->second;
    for (size_t i = 0; i < moves.size(); i++)
    {
      if (moves[i].second > env.limit)
	break;
      
      // ΤǤʤΤΤϵƤ
      const Move move = bigramHelper.indexToMove(moves[i].first, *env.state);
      if (move.isInvalid())
	continue;
      out.push_back(move, moves[i].second);
      if (!move.from().isPieceStand()
	  && move.ptype() != KING
	  && move.ptype() != GOLD
	  && !isPromoted(move.ptype())
	  && (move.from().canPromote(move.player())
	      || move.to().canPromote(move.player())))
      {
	const Move promoted_move(move.from(), move.to(),
				 promote(move.ptype()),
				 move.capturePtype(), true,
				 move.player());
	assert(env.state->isAlmostValidMove(promoted_move));
	out.push_back(promoted_move,
		      moves[i].second);
      }
    }
  }
}

int osl::category::KingNeighborBigramTable::
probability(const CategoryEnv& env, Move move, int back) const
{
  assert(back == 1 || back == 2);
  const osl::Move key_move = env.history->lastMove(back);

  if (!key_move.isNormal())
    return env.limit+1;
  const osl::Player kingTurn = back == 2 ? alt(env.state->getTurn()) : env.state->getTurn();
  const osl::Position kingPosition = env.state->getKingPosition(kingTurn);

  if (!isIn5x5(kingPosition, key_move.to())
      && !isAttackByMajorPiece(*env.state, key_move.to(), key_move.ptype(),
			       kingTurn))
    return env.limit+1;
  int index = bigramHelper.toIndex(kingPosition, key_move, env.state->getTurn());
  map_t::const_iterator p=table->find(index);
  if (p != table->end()) {
    typedef osl::stl::vector<std::pair<int, int> > vector_t;
    const vector_t& moves = p->second;
    const int value = bigramHelper.moveToValue(move);
    for (size_t i = 0; i < moves.size(); i++) {
      if (moves[i].first == value)
	return moves[i].second;
      if (moves[i].second > env.limit)
	break;
    }
  }
  return env.limit+1;
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
