/* progressEval.cc
 */
#include "osl/eval/progressEval.h"
#include "osl/container/pieceValues.h"
#include "osl/state/numEffectState.h"

BOOST_STATIC_ASSERT((osl::eval::ProgressEval::ROUND_UP 
		     & (osl::eval::ProgressEval::ROUND_UP-1))
		    == 0);

osl::eval::
ProgressEval::ProgressEval(const NumEffectState& state) 
  : opening_eval(state), endgame_eval(state),
    current_progress(state), defense_effect(state),
    minor_piece_bonus(state), cache(INVALID)
{
  pin_bonus = calculatePinBonus(state);
  mobility_bonus = calculateMobilityBonus(state);
  attack_rooks = calculateAttackRooks(state);
  assert(initialized());
}

int osl::eval::
ProgressEval::calculateMobilityBonus(const NumEffectState& state)
{
  int val=0;
  for(int i=PtypeTraits<ROOK>::indexMin;
      i<PtypeTraits<ROOK>::indexLimit;++i){
    Piece p=state.getPieceOf(i);
    if(p.isOnBoardByOwner<BLACK>()){
      int vc= RookMobility::countVerticalAll(BLACK,state,p);
      int hc= RookMobility::countHorizontalAll(BLACK,state,p);
      if(p.isPromoted()){
       val+=MobilityTable::prookVertical[vc];
       val+=MobilityTable::prookHorizontal[hc];
      }
      else{
       val+=MobilityTable::rookVertical[vc];
       val+=MobilityTable::rookHorizontal[hc];
      }
    }
    else if(p.isOnBoardByOwner<WHITE>()){
      int vc= RookMobility::countVerticalAll(WHITE,state,p);
      int hc= RookMobility::countHorizontalAll(WHITE,state,p);
      if(p.isPromoted()){
       val-=MobilityTable::prookVertical[vc];
       val-=MobilityTable::prookHorizontal[hc];
      }
      else{
       val-=MobilityTable::rookVertical[vc];
       val-=MobilityTable::rookHorizontal[hc];
      }
    }
  }
  for(int i=PtypeTraits<BISHOP>::indexMin;
      i<PtypeTraits<BISHOP>::indexLimit;++i){
    Piece p=state.getPieceOf(i);
    if(p.isOnBoardByOwner<BLACK>()){
      int c= BishopMobility::countAll(BLACK,state,p);
      if(p.isPromoted())
       val+=MobilityTable::pbishop[c];
      else
       val+=MobilityTable::bishop[c];
    }
    else if(p.isOnBoardByOwner<WHITE>()){
      int c= BishopMobility::countAll(WHITE,state,p);
      if(p.isPromoted())
       val-=MobilityTable::pbishop[c];
      else
       val-=MobilityTable::bishop[c];
    }
  }
  for(int i=PtypeTraits<LANCE>::indexMin;
      i<PtypeTraits<LANCE>::indexLimit;++i){
    Piece p=state.getPieceOf(i);
    if(p.isOnBoardByOwner<BLACK>() && !p.isPromoted()){
      int c= LanceMobility::countAll(BLACK,state,p);
      val+=MobilityTable::lance[c];
    }
    else if(p.isOnBoardByOwner<WHITE>()  && !p.isPromoted()){
      int c= LanceMobility::countAll(WHITE,state,p);
      val-=MobilityTable::lance[c];
    }
  }
  return val*128/100;
}

void osl::eval::
ProgressEval::update(const NumEffectState& new_state, Move last_move)
{
  if (last_move.isPass())
    return;
  opening_eval.update(new_state, last_move);
  endgame_eval.update(new_state, last_move);
  current_progress.update(new_state, last_move);
  defense_effect.update(new_state, last_move);
  minor_piece_bonus.update(new_state, last_move);
  pin_bonus = calculatePinBonus(new_state);
  mobility_bonus = calculateMobilityBonus(new_state);
  attack_rooks = calculateAttackRooks(new_state);
  invalidateCache();
}

int osl::eval::
ProgressEval::calculatePinBonus(const NumEffectState& state)
{
  int bonus = 0;
  const Piece black_king = state.getKingPiece<BLACK>();
  const Piece white_king = state.getKingPiece<WHITE>();
  osl::container::PieceMask white_mask =
    pin_mask[playerToIndex(WHITE)] =
    osl::effect_util::Pin::make(state, WHITE);
  while (white_mask.any())
  {
    const Piece piece = state.getPieceOf(white_mask.takeOneBit());
    bonus -= endgame_eval.valueOf(
      black_king, white_king,
      piece) / 4;
  }

  osl::container::PieceMask black_mask =
    pin_mask[playerToIndex(BLACK)] = 
    osl::effect_util::Pin::make(state, BLACK);
  while (black_mask.any())
  {
    const Piece piece = state.getPieceOf(black_mask.takeOneBit());
    bonus -= endgame_eval.valueOf(
      black_king, white_king,
      piece) / 4;
  }
	
  return bonus;
}

int osl::eval::
ProgressEval::calculateAttackRooks(const NumEffectState& state)
{
  int rooks = 0;
  for(int i = PtypeTraits<ROOK>::indexMin;
      i < PtypeTraits<ROOK>::indexLimit; ++i)
  {
    const Piece rook = state.getPieceOf(i);
    if (rook.isOnBoard() && rook.position().canPromote(rook.owner()) &&
	state.getKingPiece(alt(rook.owner())).position().canPromote(rook.owner()))
    {
      if (rook.owner() == BLACK)
	++rooks;
      else
	--rooks;
    }
  }
  return rooks;
}

void osl::eval::
ProgressEval::setValues(const SimpleState& state, Progress16 progress16,
			PieceValues& out)
{
  PieceValues opening, endgame;
  const NumEffectState nstate(state);
  const progress_t progress(nstate);
  const defense_t defense_effect(nstate);
  const MinorPieceBonus minor_piece_bonus(state);
  opening_eval_t::setValues(state, opening);
  endgame_eval_t::setValues(state, endgame);
  for (int i=0; i<Piece::SIZE; ++i)
  {
    out[i] = composeValue(opening[i] & (~1), endgame[i] & (~1), progress16,
			  progress.progress16(BLACK),
			  progress.progress16(WHITE),
			  defense_effect.progress16(BLACK),
			  defense_effect.progress16(WHITE),
			  minor_piece_bonus.value(progress16), 0, 0, 0);
  }
}

void osl::eval::
ProgressEval::setValues(const SimpleState& state, PieceValues& out)
{
  const NumEffectState nstate(state);
  const progress_t progress(nstate);
  setValues(state, progress.progress16(), out);
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
