/* gameState.cc
 */
#include "osl/game_playing/gameState.h"
#include "osl/game_playing/openingBookTracer.h"
#include "osl/move_classifier/pawnDropCheckmate.h"
#include "osl/move_classifier/moveAdaptor.h"
#include "osl/effect_util/effectUtil.h"
#include "osl/apply_move/doUndoMoveStack.h"
#include "osl/hash/hashKeyStack.h"
#include "osl/container/moveStack.h"
#include "osl/repetitionCounter.h"
#include "osl/sennichite.h"

struct osl::game_playing::GameState::State
{
  typedef DoUndoMoveStack<HashEffectState> lock_stack_t;
  HashEffectState state;
  RepetitionCounter counter;
  lock_stack_t move_stack;
  MoveStack move_history;
  const SimpleState initial_state;
  
  State(const SimpleState& initial_state) 
    : state(initial_state), counter(state), initial_state(initial_state)
  {
    move_history.reserve(1024);
  }
  enum CopyType { FORK };	// forkʤ
  State(CopyType, const State& src)
    : state(src.state), counter(src.counter), move_history(src.move_history),
      initial_state(src.initial_state)
  {
  }
};

osl::game_playing::
GameState::GameState(const SimpleState& initial_state) 
  : stack(new State(initial_state))
{
}

osl::game_playing::
GameState::GameState(const State& src) 
  : stack(new State(State::FORK, src))
{
}

osl::game_playing::
GameState::~GameState()
{
}

const osl::Sennichite osl::game_playing::
GameState::pushMove(Move m)
{
  stack->move_history.push(m);
  const Sennichite result
    = stack->counter.isSennichite(stack->state, m);
  stack->counter.push(stack->state, m);
  stack->move_stack.push(stack->state, m);
  return result;
}

osl::game_playing::GameState::MoveType osl::game_playing::
GameState::isIllegal(Move m) const
{
  if (! state().isValidMove(m, false))
    return OTHER_INVALID;
  typedef move_classifier::PlayerMoveAdaptor<move_classifier::PawnDropCheckmate>
    PawnDropCheckmate_t;
  if (PawnDropCheckmate_t::isMember(state(), m))
    return PAWN_DROP_FOUL;

  stack->move_stack.push(stack->state, m);
  const bool unsafe_king
    = EffectUtil::isKingInCheck(alt(state().getTurn()), state());
  stack->move_stack.pop();
  
  if (unsafe_king)
    return UNSAFE_KING;
  
  return VALID;
}

const osl::Move osl::game_playing::
GameState::popMove()
{
  const Move result = stack->move_history.lastMove();
  assert(canPopMove());
  stack->move_history.pop();
  stack->counter.pop();
  stack->move_stack.pop();
  return result;
}

const osl::HashEffectState& osl::game_playing::
GameState::state() const
{
  return stack->state; 
}

int osl::game_playing::
GameState::moves() const 
{ 
  return stack->move_history.size();
}

const osl::MoveStack& osl::game_playing::
GameState::moveHistory() const 
{ 
  return stack->move_history;
}

const osl::hash::HashKeyStack& osl::game_playing::
GameState::hashHistory() const
{
  return stack->counter.history();
}

const osl::RepetitionCounter& osl::game_playing::
GameState::counter() const
{
  return stack->counter;
}

bool osl::game_playing::
GameState::canPopMove() const
{
  return ! stack->move_stack.empty();
}

const boost::shared_ptr<osl::game_playing::GameState> osl::game_playing::
GameState::clone() const
{
  boost::shared_ptr<GameState> result(new GameState(*stack));
  return result;
}

const osl::state::SimpleState& osl::game_playing::
GameState::getInitialState() const
{
  return stack->initial_state;
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
