/* threadForEachMove.cc
 */
#include "osl/game_playing/threadForEachMove.h"
#include "osl/game_playing/gameState.h"
#include "osl/game_playing/searchPlayer.h"
#include "osl/search/timeControl.h"
#include "osl/search/searchRecorder.h"
#include "osl/misc/realTime.h"
#include "osl/record/csa.h"
#include "osl/sennichite.h"
#include <boost/thread.hpp>
#include <exception>
#include <iostream>
#include <ctime>
#include <unistd.h>

#define DEBUG_SPECULATIVE_EXECUTION
#define EXTRA_DEBUG_SPECULATIVE_EXECUTION

struct osl::game_playing::ThreadForEachMove::SpeculativeThread
{
  boost::shared_ptr<boost::thread> thread;
  boost::shared_ptr<GameState> state;
  boost::shared_ptr<SearchPlayer> player;
  bool parent_is_verbose;
  volatile bool thread_finished;
  Move predicted_move;
  MoveWithComment next_move;
  bool verbose;
  // operator()񤭹
  MoveWithComment *next_move_for_thread;
  volatile bool *thread_finished_for_thread;

  SpeculativeThread() : player(), parent_is_verbose(false),
			thread_finished(true),
			verbose(false),
			next_move_for_thread(&next_move),
			thread_finished_for_thread(&thread_finished)
  {
  }
private:
  void errorInThread()
  {
    std::cerr << "caught exception in thread" << std::cerr;
    *next_move_for_thread = MoveWithComment(Move::INVALID());
    *thread_finished_for_thread = true;
  }
public:
  void operator()()
  {
    // boost::thread ˤϤobject *copy* 
    // ϤΤǡǤϥСѿɤ߽ФǤʤ
    // i.e., ͤѹƤФˡʤ
    // äƽ񤯻ˤϥݥ󥿷ͳǡ
    assert(predicted_move.isNormal());
    *thread_finished_for_thread = false;
    try
    {
      state->pushMove(predicted_move);
      player->pushMove(predicted_move);
      player->setVerbose(verbose);
      *next_move_for_thread = MoveWithComment(Move::INVALID());
      *next_move_for_thread = player->selectBestMove(*state, -1, -1); // infinity
    }
    catch (std::exception& e)
    {
      errorInThread();
      std::cerr << e.what() << std::endl;
      return;
    }
    catch (...)
    {
      errorInThread();
      std::cerr << "unknown exception" << std::endl;
      return;
    }
    *thread_finished_for_thread = true;
#ifdef DEBUG_SPECULATIVE_EXECUTION
    time_t now = time(0);
    char ctime_buf[64];
    std::cerr << "if " << record::csa::show(predicted_move) 
	      << " then " << record::csa::show(next_move_for_thread->move)
	      << " " << ctime_r(&now, ctime_buf);
#endif
    if (parent_is_verbose)
      player->recorder().reportCount(std::cerr);
  }
  void run(Move prediction, const GameState& main_state,
	   const SearchPlayer& main_player, bool verbose);
private:
  void spawnFailed()
  {
    char ctime_buf[64];
    time_t now = time(0);
    std::cerr << "thread creation failure " << ctime_r(&now, ctime_buf);
    thread.reset();
    player.reset();
    predicted_move = Move::INVALID();
  }
public:
  void join()
  {
    if (thread.get())
    {
      thread->join();
      thread.reset();
    }
    predicted_move = Move::INVALID();
  }
  void waitRunning() const
  {
    if (thread_finished)
      return;
    while ((! thread_finished)
	   && (! player->canStopSearch()))
    {
      timespec rq = { 0, 20000000 }, rm;
      while (nanosleep(&rq, &rm) == -1) 
      {
	rq = rm;
      }
    }
  }
  void changeTimer(const misc::RealTime& timer)
  {
    waitRunning();
    if (! thread_finished)
      player->setStopSchedule(timer);
  }
  void stopSearchNow()
  {
    char ctime_buf[64];
    if (! thread_finished)
    {
#ifdef EXTRA_DEBUG_SPECULATIVE_EXECUTION
      time_t now = time(0);
      std::cerr << "need to call stopSearchNow " << ctime_r(&now, ctime_buf);
#endif
      waitRunning();
      if (! thread_finished)
      {
#ifdef EXTRA_DEBUG_SPECULATIVE_EXECUTION
	time_t now = time(0);
	std::cerr << "calling stopSearchNow " << ctime_r(&now, ctime_buf);
#endif
	const bool success
	  = player->stopSearchNow();
	if (! success)
	{
	  std::cerr << "stop search failed" << this << "\n";
	}
      }
    }
    join();
  }
};

void osl::game_playing::ThreadForEachMove::SpeculativeThread::
run(Move prediction, const GameState& main_state,
    const SearchPlayer& main_player, bool verbose)  
{
  try
  {
    player.reset();
    player.reset(dynamic_cast<SearchPlayer*>(main_player.clone()));
    parent_is_verbose = player->getConfig().verbose;
    assert(! thread.get());
    predicted_move = prediction;
    state = main_state.clone();
    this->verbose = verbose;
    thread.reset(new boost::thread(*this));
#ifdef DEBUG_SPECULATIVE_EXECUTION
    char ctime_buf[64];
    time_t now = time(0);
    std::cerr << "speculative search for " << record::csa::show(predicted_move) 
	      << " " << ctime_r(&now, ctime_buf);
#endif
  }
  catch (boost::thread_resource_error&)
  {
    spawnFailed();
  }
  catch (std::exception& e)
  {
    spawnFailed();
    std::cerr << e.what() << std::endl;
  }
  catch (...)
  {
    spawnFailed();
    std::cerr << "unknown exception" << std::endl;
  }
}

osl::game_playing::ThreadForEachMove::
ThreadForEachMove(int threads)
  : speculative_thread0(new SpeculativeThread()),
    speculative_thread1(new SpeculativeThread()),
    max_threads(threads)
{
}
osl::game_playing::
ThreadForEachMove::~ThreadForEachMove()
{
  stopAll();
  selectBestMoveCleanUp();
}

void osl::game_playing::ThreadForEachMove::
startSpeculative(const boost::shared_ptr<GameState> state,
		 const SearchPlayer& main_player)
{
  assert(speculative_thread0.get());
  assert(speculative_thread1.get());
  MoveVector moves;
  main_player.pickUpBestMoves(*state, moves);
  const bool run_thread0 = moves.size() > 0 && max_threads > 0;
  const bool run_thread1 = moves.size() > 1 && max_threads > 1;
  if (run_thread0)
    speculative_thread0->run(moves[0], *state, main_player,
			     (! run_thread1));
  if (run_thread1)
    speculative_thread1->run(moves[1], *state, main_player,
			     false);
}

void osl::game_playing::ThreadForEachMove::
stopOtherThan(Move m)
{
  char ctime_buf[64];
  if (speculative_thread0->predicted_move != m)
  {
#ifdef EXTRA_DEBUG_SPECULATIVE_EXECUTION
    if (max_threads > 0)
    {
      time_t now = time(0);
      std::cerr << "stop thread0 " << ctime_r(&now, ctime_buf);
    }
#endif
    speculative_thread0->stopSearchNow();
  }
  if (speculative_thread1->predicted_move != m)
  {
#ifdef EXTRA_DEBUG_SPECULATIVE_EXECUTION
    if (max_threads > 1)
    {
      time_t now = time(0);
      std::cerr << "stop thread1 " << ctime_r(&now, ctime_buf);
    }
#endif
    speculative_thread1->stopSearchNow();
  }
}

void osl::game_playing::ThreadForEachMove::
stopAll()
{
  if (! speculative_thread0->thread_finished)
    speculative_thread0->stopSearchNow();
  if (! speculative_thread1->thread_finished)
    speculative_thread1->stopSearchNow();
}

const osl::game_playing::MoveWithComment
osl::game_playing::ThreadForEachMove::
waitResult(Move last_move, const int wait_for,
	   SearchPlayer& main_player, int /*byoyomi*/)
{
  const time_t start_time = time(0);
  assert(last_move.isNormal());
  SpeculativeThread *match = 0;
  MoveWithComment result = MoveWithComment(Move::INVALID());

  if (speculative_thread0->predicted_move == last_move)
    match = speculative_thread0.get();
  else if (speculative_thread1->predicted_move == last_move)
    match = speculative_thread1.get();
  
  if (! match)
  {
#ifdef DEBUG_SPECULATIVE_EXECUTION
    time_t now = time(0);
    char ctime_buf[64];
    std::cerr << "unexpected " << record::csa::show(last_move)
	      << " " << ctime_r(&now, ctime_buf);
#endif    
    return result;
  }
  
  const time_t stop_time = start_time + wait_for;
  match->changeTimer(misc::RealTime(wait_for));
#ifdef DEBUG_SPECULATIVE_EXECUTION
  char ctime_buf[64];
  std::cerr << "match " << ctime_r(&start_time, ctime_buf);
  std::cerr << "wait for " << ctime_r(&stop_time, ctime_buf);
#endif
  match->player->setVerbose(main_player.getConfig().verbose);
  while (true)
  {
    if (match->thread_finished)
      break;

    time_t now = time(0);
    if (now >= stop_time)
    {
      match->stopSearchNow();
#ifdef EXTRA_DEBUG_SPECULATIVE_EXECUTION
      std::cerr << "stop search (match): " << ctime_r(&now, ctime_buf);
#endif
      break;
    }
    timespec rq = { 1, 0 }, rm;
    nanosleep(&rq, &rm);
  }
  result = match->next_move;
  if (result.move.isNormal()) {
    main_player.swapTable(*match->player);
  }
  return result;
}
  
void osl::game_playing::ThreadForEachMove::
selectBestMoveCleanUp()
{
  speculative_thread0->join();
  speculative_thread1->join();
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
