/* nonBlockDelete.cc
 */
#include "osl/misc/nonBlockDelete.h"
#include <boost/thread.hpp>
#include <deque>
#include <iostream>

class osl::misc::NonBlockDelete::Queue
{
  typedef std::deque<boost::shared_ptr<void> > queue_t;
  queue_t data;
public:
  Queue() : finish(false)
  {
  }
  boost::condition condition;
  typedef boost::mutex Mutex;
  mutable Mutex mutex;
  volatile bool finish;

  size_t size() const
  {
    Mutex::scoped_lock lk(mutex);
    return data.size();
  }
  void push_back(boost::shared_ptr<void>& ptr)
  {
    Mutex::scoped_lock lk(mutex);
    data.push_back(boost::shared_ptr<void>());
    data.back().swap(ptr);
    condition.notify_one();
  }
private:
  boost::shared_ptr<void> pop_front_in_lock()
  {
    if (! data.empty())
    {
      boost::shared_ptr<void> result = data.front();
      data.pop_front();
      return result;
    }
    return boost::shared_ptr<void>();
  }
public:
  boost::shared_ptr<void> pop_front_non_block()
  {
    Mutex::scoped_lock lk(mutex);
    return pop_front_in_lock();
  }
  boost::shared_ptr<void> pop_front()
  {
    while (! finish) 
    {
      Mutex::scoped_lock lk(mutex);
      boost::shared_ptr<void> result = pop_front_in_lock();
      if (result.get() || finish)
	return result;
      condition.wait(lk);
    }
    return boost::shared_ptr<void>();
  }
};

struct osl::misc::NonBlockDelete::Runner
{
  boost::shared_ptr<Queue> queue;
  Runner(const boost::shared_ptr<Queue>& q) : queue(q)
  {
  }
  void operator()()
  {
    while (! queue->finish)
    {
      boost::shared_ptr<void> ptr;
      ptr = queue->pop_front();
      const int count = ptr.use_count();
      if (count > 1)
	std::cerr << "NonBlockDelete " << count << " > 1 " 
		  << ptr.get() << std::endl;
      // release ptr
    }
  }
};


osl::misc::NonBlockDelete& osl::misc::
NonBlockDelete::instance()
{
  static NonBlockDelete the_instance;
  return the_instance;
}

osl::misc::
NonBlockDelete::NonBlockDelete()
  : queue(new Queue())
{
  boost::thread(Runner(queue));
}

osl::misc::
NonBlockDelete::~NonBlockDelete()
{
  queue->finish = true;
}

int osl::misc::
NonBlockDelete::waiting()
{
  NonBlockDelete& the_instance = instance();
  return the_instance.queue->size();
}

void osl::misc::NonBlockDelete::resetAny(boost::shared_ptr<void>& ptr)
{
  instance().push_back(ptr);
}

void osl::misc::NonBlockDelete::push_back(boost::shared_ptr<void>& ptr)
{
  queue->push_back(ptr);
}

void osl::misc::
NonBlockDelete::deleteAll()
{
  NonBlockDelete& the_instance = instance();
  while (the_instance.queue->size())
  {
    boost::shared_ptr<void> ptr
      = the_instance.queue->pop_front_non_block();
  }
  boost::thread::yield();
}


/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
