/* featureSet.cc
 */
#include "osl/rating/featureSet.h"
#include "osl/rating/group.h"
#include "osl/rating/group/captureGroup.h"
#include "osl/rating/group/escape.h"
#include "osl/rating/group/positionGroup.h"
#include "osl/rating/group/patternGroup.h"
#include "osl/rating/group/bigramGroup.h"
#include "osl/rating/feature/checkmate.h"
#include "osl/move_generator/legalMoves.h"
#include "osl/container/moveVector.h"
#include "osl/oslConfig.h"
#include <iostream>
#include <sstream>
#include <iomanip>
#include <stdexcept>
#include <cmath>

osl::rating::
FeatureSet::FeatureSet()
  : capture_group(-1), checkmate_if_capture_group(-1)
{
}

osl::rating::
FeatureSet::~FeatureSet()
{
}

const osl::rating::range_t osl::rating::
FeatureSet::makeRange(size_t group) const
{
  int first = 0;
  for (size_t i=0; i<groups.size(); ++i) {
    if (i == group)
      return std::make_pair(first, first+groups[i].size());
    first += groups[i].size();
  }
  assert(0);
  abort();
}

void osl::rating::
FeatureSet::addFinished()
{
  assert(normal_groups.size() == groups.size());
  weights.resize(features.size(), 1.0);
  weightslog10.resize(features.size(), 1);
  assert(weights.size() == features.size());
  ranges.resize(groups.size());
  for (size_t i=0; i<groups.size(); ++i)
    ranges[i] = makeRange(i);
}

void osl::rating::
FeatureSet::add(Feature *f)
{
  add(new Group(f));
}

void osl::rating::
FeatureSet::addCommon(Group *g)
{
  for (size_t i=0; i<g->size(); ++i) {
    features.push_back(&(*g)[i]);
  }
  groups.push_back(g);
}

void osl::rating::
FeatureSet::add(Group *g)
{
  normal_groups.push_back(true);
  addCommon(g);
}

void osl::rating::
FeatureSet::add(CaptureGroup *g)
{
  capture_group = normal_groups.size();
  normal_groups.push_back(false);
  addCommon(g);
}

void osl::rating::
FeatureSet::add(CheckmateIfCapture *f)
{
  checkmate_if_capture_group = normal_groups.size();
  normal_groups.push_back(false);  
  addCommon(new Group(f));
}

bool osl::rating::
FeatureSet::tryLoad(const std::string& input_directory) 
{
  bool result = true;
  for (size_t i=0; i<groups.size(); ++i) {
    const bool success = groups[i].load(input_directory, ranges[i], weights);
    if (! success && result)
      std::cerr << "warning: rating load failed " << groups[i].group_name << " " << i 
		<< " in " << input_directory << "\n";
    result &= success;
  }
  for (size_t i=0; i<features.size(); ++i)
    weightslog10[i] = static_cast<int>(400*log10(weights[i]));
  return result;
}

void osl::rating::
FeatureSet::setWeight(size_t feature_id, const double& value)
{
  weights[feature_id] = value; 
  weightslog10[feature_id] = static_cast<int>(400*log10(value));
}

void osl::rating::
FeatureSet::generateRating(const NumEffectState& state, const RatingEnv& env,
			   RatedMoveVector& out) const
{
  MoveVector moves;
  LegalMoves::generate(state, moves);
  for (size_t i=0; i<moves.size(); ++i) {
    out.push_back(makeRate(state, env, moves[i]));
  }
  out.sort();
}

// width 4, [0,80]
static const osl::CArray<int, 20> order_to_depth = {{
  // 211, 256, 313, 335, 360, 375, 386, 403, 408, 449,
  200, 250, 300, 340, 360, 380, 390, 400, 410, 420,
  420, 450, 450, 450, 450, 450, 450, 450, 460, 460, 
}};
static const osl::CArray<int, 20> order_to_width = {{
  // 273, 435, 561, 638, 712, 769, 819, 862, 883,1063,
  270, 440, 550, 640, 710, 770, 820, 860, 880, 910,
  915, 940,1000,1000,1015,1060,1060,1060,1100,1200,
}};

const int sc_width = 100, sc_length = 12, sc_start = -500;
static const osl::CArray<int, sc_length> score_to_logprob = {{
  //  -500 -  -400      378
  //  -400 -  -300      378
  //  -300 -  -200      380
  //  -200 -  -100      364
  //  -100 -     0      352
  //     0 -   100      327
  //   100 -   200      294
  //   200 -   300      291
  //   300 -   400      279
  //   400 -   500      247
  //   500 -   600      213
  //   600 -   700      146
  380, 380, 380, 360, 350, 330, 290, 290, 280, 250, 210, 150,
}};
static const osl::CArray<int, sc_length> score_to_logprob_limit = {{
  // 1203, 977, 883, 783, 681, 573, 474, 446, 421, 370, 320, 207,
  1200, 980, 880, 780, 680, 580, 470, 450, 420, 370, 320, 200,
}};

static const osl::CArray<int, sc_length> relative_score_to_logprob = {{
  // 232, 254, 287, 310, 328, 354, 371, 395, 402, 410, 421, 436,
  230, 250, 280, 310, 330, 350, 370, 400, 400, 410, 420, 430
}};
static const osl::CArray<int, sc_length> relative_score_to_logprob_limit = {{
  // 249, 331, 416, 492, 550, 619, 682, 750, 795, 848, 905, 1084,
  250, 330, 410, 490, 550, 620, 680, 750, 790, 850, 900, 1080,
}};

inline int make_prob(int score, int order, int limit, int highest) 
{
  const int order_index = std::min((int)order/4, 19);
  int result = limit+1;
  if (order_to_width[order_index] <= limit) {
    result = (order == 0) ? 100 : order_to_depth[order_index];
  }  
  score = std::max(sc_start, score);
  const int score_index = std::min((score - sc_start)/sc_width, sc_length-1);
  if (score_to_logprob_limit[score_index] <= limit) {
    result = std::min(result, score_to_logprob[score_index]);
  }
  if (order > 0) {
    const int rscore_index = std::min((highest - score)/100, sc_length-1);
    assert(rscore_index >= 0);
    if (relative_score_to_logprob_limit[rscore_index] <= limit)
      result = std::min(result, relative_score_to_logprob[rscore_index]);
  }
  return result;
}


void osl::rating::
FeatureSet::generateLogProb(const NumEffectState& state, const RatingEnv& env,
			    int limit, MoveLogProbVector& out) const
{
  RatedMoveVector score;
  generateRating(state, env, score);
  if (score.empty())
    return;
  const int highest = score[0].rating();
  for (size_t i=0; i<score.size(); ++i) {
    out.push_back(MoveLogProb(score[i].move(), make_prob(score[i].rating(), i, limit, highest)));
  }
}

const int max_score = 999, min_score = 0;
static const osl::CArray<int, 10> score_to_logprob_takeback = {{
  // 184, 280, 243, 176, 158, 131,  71,  74,  92,  64,
  250, 250, 180, 160, 130,  80,  80,  80,  65, 50
}};
static const osl::CArray<int, 10> score_to_logprob_seeplus = {{
  // 408, 444, 289, 287, 246, 203, 179, 190, 145, 127,
  450, 300, 300, 250, 200, 180, 180, 150, 130, 100,
}};
static const osl::CArray<int, 10> score_to_logprob_kingescape = {{
  // 75,  97,  97,  35, 100, 127,  47,  79,  48,  57, 
  100, 100, 100, 100, 100, 100, 80, 80, 50, 50,
}};

int osl::rating::
FeatureSet::logProbTakeBack(const NumEffectState& state, const RatingEnv& env, Move move) const
{
  const int score = makeRate(state, env, move).rating();
  return score_to_logprob_takeback[std::max(min_score, std::min(max_score, score))/100];
}
int osl::rating::
FeatureSet::logProbSeePlus(const NumEffectState& state, const RatingEnv& env,
			   Move move) const
{
  const int score = makeRate(state, env, move).rating();
  return score_to_logprob_seeplus[std::max(min_score, std::min(max_score, score))/100];
}
int osl::rating::
FeatureSet::logProbKingEscape(const NumEffectState& state, const RatingEnv& env,
			      Move move) const
{
  const int score = makeRate(state, env, move).rating();
  const int prob = score_to_logprob_kingescape[std::max(min_score, std::min(max_score, score))/100];
  assert(prob > 0);
  return prob;
}

int osl::rating::
FeatureSet::rating(const NumEffectState& state, 
		   const RatingEnv& env, Move move, size_t group_id) const
{
  int found = groups[group_id].findMatch(state, move, env);
  if (found < 0)
    return 0;
  found += ranges[group_id].first;
  return weightslog10[found];
}

const osl::rating::RatedMove osl::rating::
FeatureSet::makeRate(const NumEffectState& state, 
		     const RatingEnv& env, Move move) const
{
  int sum = 0;
  for (size_t j=0; j<groups.size(); ++j) {
    if (! normal_groups[j])
      continue;
    sum += rating(state, env, move, j);
  }
  int capture = 0;
  if (capture_group >= 0)
    capture = rating(state, env, move, capture_group);
  int checkmate_if_capture = 0;
  if (checkmate_if_capture_group >= 0)
    checkmate_if_capture = rating(state, env, move, checkmate_if_capture_group);
  sum += checkmate_if_capture;

  if (checkmate_if_capture > 0)
    capture = std::max(0, capture);
  const int optimistic = sum + std::max(0, capture);
  sum += capture;
  
  return RatedMove(move, sum, optimistic);
}

const std::string osl::rating::
FeatureSet::annotate(const NumEffectState& state, 
		     const RatingEnv& env, Move move) const
{
  vector<std::pair<int, std::string> > values;
  for (size_t j=0; j<groups.size(); ++j) {
    int found = groups[j].findMatch(state, move, env);
    if (found < 0)
      continue;
    found += ranges[j].first;
    values.push_back(std::make_pair(weightslog10[found], groups[j].group_name));
  }
  std::sort(values.begin(), values.end());
  std::reverse(values.begin(), values.end());
  std::ostringstream ss;
  for (size_t i=0; i<values.size(); ++i) {
    if (i)
      ss << "  ";
    ss << values[i].second << " " << values[i].first;
  }
  return ss.str();
}

void osl::rating::
FeatureSet::showGroup(std::ostream& os, size_t group_id) const
{
  os << std::setprecision(3);
  group(group_id).show(os, 12, range(group_id), weights);
}

void osl::rating::
FeatureSet::save(const std::string& output_directory, size_t group_id) const
{
  group(group_id).saveResult(output_directory, range(group_id), weights);
}

/* ------------------------------------------------------------------------- */

osl::rating::
StandardFeatureSet::StandardFeatureSet(bool allow_load_failure)
{
  add(new TakeBackGroup());
  add(new CaptureGroup());
  add(new ContinueCapture());
  add(new AttackPin(true));
  add(new AttackPin(false));
  add(new Karanari());
  add(new RookDefense());
  add(new BadLance());
  add(new CheckGroup());
  add(new CheckmateIfCapture());
  add(new Threatmate());
  add(new SendOff());
  add(new FromEffectGroup());
  add(new PtypeAttackedGroup());
  add(new ToSupported());
  add(new ShadowEffectGroup());
  add(new BlockGroup());
  add(new OpenGroup());

  add(new RelativeKingXGroup(true));
  add(new RelativeKingYGroup(true));
  add(new RelativeKingXGroup(false));
  add(new RelativeKingYGroup(false));
  add(new PositionXGroup());
  add(new PositionYGroup());
  
  add(new PatternGroup(U));
  add(new PatternGroup(UL));
  add(new PatternGroup(UR));
  add(new PatternGroup(L));
  add(new PatternGroup(R));
  add(new PatternGroup(D));
  add(new PatternGroup(DL));
  add(new PatternGroup(DR));

  add(new PatternGroup(UUL));
  add(new PatternGroup(UUR));

  add(new PatternLongGroup(0));
  add(new PatternLongGroup(1));
  add(new PatternLongGroup(2));
  add(new PatternLongGroup(3));

  add(new PatternLongGroup2(0));
  add(new PatternLongGroup2(1));
  add(new PatternLongGroup2(2));
  add(new PatternLongGroup2(3));

  add(new BigramAttackGroup(true, true));
  add(new BigramAttackGroup(true, false));
  add(new BigramAttackGroup(false, true));
  add(new BigramAttackGroup(false, false));

  addFinished();
  bool success = tryLoad(OslConfig::home()+"/data/rating");
  if (! allow_load_failure && ! success) {
    std::cerr << "error: unable to load rating from " << OslConfig::home()+"/data/rating";
    throw std::runtime_error("load failed " + OslConfig::home()+"/data/rating");
  }
}

osl::rating::
CaptureSet::CaptureSet(bool allow_load_failure)
{
  add(new CaptureGroup());
  add(new ShadowEffectGroup());

  addFinished();
  bool success = tryLoad(OslConfig::home()+"/data/rating");
  if (! allow_load_failure && ! success) {
    std::cerr << "error: unable to load rating from " << OslConfig::home()+"/data/rating";
    throw std::runtime_error("load failed " + OslConfig::home()+"/data/rating");
  }
}

osl::rating::
TacticalSet::TacticalSet(bool allow_load_failure)
{
  add(new TakeBackGroup());
  add(new CaptureGroup());
  add(new ContinueCapture());
  add(new AttackPin(true));
  add(new AttackPin(false));
  add(new CheckGroup());
  add(new CheckmateIfCapture());
  add(new Threatmate());
  add(new SendOff());
  add(new FromEffectGroup());
  add(new PtypeAttackedGroup());
  add(new ToSupported());
  add(new ShadowEffectGroup());
  add(new BlockGroup());
  add(new OpenGroup());

  addFinished();
  bool success = tryLoad(OslConfig::home()+"/data/rating");
  if (! allow_load_failure && ! success) {
    std::cerr << "error: unable to load rating from " << OslConfig::home()+"/data/rating";
    throw std::runtime_error("load failed " + OslConfig::home()+"/data/rating");
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
