/* alphaBeta2.cc
 */
#include "osl/search/alphaBeta2.h"
#ifdef OSL_SMP
#  include "osl/search/alphaBeta2Parallel.h"
#endif
#include "osl/search/simpleHashRecord.h"
#include "osl/search/searchMoveVector.h"
#include "osl/search/simpleHashTable.h"
#include "osl/search/dominanceCheck.h"
#include "osl/search/moveGenerator.h"
#include "osl/search/realizationProbability.h"
#include "osl/search/quiescenceSearch2.h"
#include "osl/search/realizationProbability.h"
#include "osl/checkmate/immediateCheckmate.h"
#include "osl/record/csa.h"
#include "osl/move_classifier/pawnDropCheckmate.h"
#include "osl/move_classifier/check_.h"
#include "osl/move_classifier/moveAdaptor.h"
#include "osl/effect_util/additionalEffect.h"
#include "osl/misc/nonBlockDelete.h"
#include <stdexcept>
#include <iostream>
#include <iomanip>

#define search_assert(x, m) assert((x) || SearchState2::abort(m))

typedef osl::search::RealizationProbability Probabilities_t;

#ifdef CHECKMATE_COUNT
static size_t root_checkmate = 0, checkmate_before = 0, checkmate_after = 0,
  count_threatmate = 0, quiesce_checkmate=0;
#endif

/* ------------------------------------------------------------------------- */
/*	Constructors							     */
/* ------------------------------------------------------------------------- */

osl::CArray<int, osl::search::AlphaBeta2Tree::MaxDepth> osl::search::AlphaBeta2Tree::depth_node_count;

osl::search::AlphaBeta2Tree::
AlphaBeta2Tree(const HashEffectState& s, checkmate_t& c, 
	       SimpleHashTable *t, CountRecorder& r)
  : SearchBase<eval::ProgressEval,SimpleHashTable,CountRecorder,RealizationProbability>(r, t), 
    SearchState2(s, c), AlphaBeta2Common(s), node_count(0),
    stop_by_alarm(Alarm::UNAVAILABLE), has_stop_alarm(false)
{
#ifdef OSL_SMP
  shared.reset(new AlphaBeta2Parallel(this));
#endif
}

osl::search::AlphaBeta2Tree::
AlphaBeta2Tree(const AlphaBeta2Tree& src)
  : SearchBase<eval::ProgressEval,SimpleHashTable,CountRecorder,RealizationProbability>(src), 
    SearchState2(src), HasTimer(src), AlphaBeta2Common(src), 
    node_count(0), stop_by_alarm(Alarm::UNAVAILABLE), has_stop_alarm(false)
{
  // copy outside of search
#ifdef OSL_SMP
  shared.reset(new AlphaBeta2Parallel(this));
#endif
  assert(src.blocking == 0);
  for (size_t i=0; i<src.generators.size(); ++i) {
    generators.push_back(alloc());
    *generators[i] = *src.generators[i];
  }
}

osl::search::AlphaBeta2Tree::
AlphaBeta2Tree(const AlphaBeta2Tree& src, AlphaBeta2Parallel *)
  : SearchBase<eval::ProgressEval,SimpleHashTable,CountRecorder,RealizationProbability>(src), 
    SearchState2(src), HasTimer(src), AlphaBeta2Common(src), 
    node_count(0), shared(src.shared), stop_by_alarm(Alarm::UNAVAILABLE), has_stop_alarm(false)
{
}

osl::search::AlphaBeta2Tree::
~AlphaBeta2Tree()
{
  for (size_t i=0; i<generators.size(); ++i)
    dealloc(generators[i]);
#if 0 // def OSL_SMP
  if (shared.use_count() == 1)
    NonBlockDelete::reset(shared);
#endif
}

osl::search::MoveGenerator *
osl::search::AlphaBeta2Tree::alloc()
{
  return new MoveGenerator;
}

void osl::search::AlphaBeta2Tree::dealloc(MoveGenerator *p)
{
  delete p;
}

osl::search::MoveGenerator& osl::search::AlphaBeta2Tree::makeGenerator()
{
  const size_t cur_depth = curDepth();
  while (generators.size() <= cur_depth)
    generators.push_back(0);
  if (generators[cur_depth] == 0)
    generators[cur_depth] = alloc();
  return *generators[cur_depth];
}

volatile int& osl::search::
AlphaBeta2Tree::stopByAlarm()
{
#ifdef OSL_SMP
  return shared->stop_by_alarm;
#else
  return stop_by_alarm;
#endif
}

const volatile int& osl::search::
AlphaBeta2Tree::stopByAlarm() const
{
#ifdef OSL_SMP
  return shared->stop_by_alarm;
#else
  return stop_by_alarm;
#endif
}

void osl::search::
AlphaBeta2Tree::setTimeOut(int seconds)
{
  if (seconds <= 0)
  {
    if (seconds < 0)
      std::cerr << "warning: setTimeOut negative " << seconds << "\n";
    seconds = 1;
  }
  has_stop_alarm = Alarm::set(seconds, AlarmSwitch(&stopByAlarm()));
  if (! has_stop_alarm)
  {
    std::cerr << "warning: search without alarm\n";
  }
}
void osl::search::
AlphaBeta2Tree::resetTimeOut()
{
  if (has_stop_alarm)
    Alarm::reset(AlarmSwitch(&stopByAlarm()));
}

void osl::search::
AlphaBeta2Tree::throwStop()
{
  if (stopByAlarm() == Alarm::TIMEOUT) {
#ifdef DEBUG
    const time_t now = time(0);
    char ctime_buf[64];
    std::cerr << "throwTimeOut " << ctime_r(&now, ctime_buf);
#endif
    resetTimeOut();
    throw misc::NoMoreTime();
  }
  assert(stop);
  throw BetaCut();
}

/* ------------------------------------------------------------------------- */
/*	Methods for alpha beta search					     */
/* ------------------------------------------------------------------------- */

template <osl::Player P>
int osl::search::AlphaBeta2Tree::
alphaBetaSearchAfterMove(const MoveLogProb& moved, Window w,
			 bool in_pv)
{
  assert(w.alpha(P) % 2);
  assert(w.beta(P) % 2);
  assert(alt(P) == state().getTurn());
  assert(P == moved.getMove().player());
  assert(eval::notLessThan(P, w.beta(P), w.alpha(P)));

  // PؤPβ => ľμ꤬ˡ
  if (EffectUtil::isKingInCheck(P, state())) {
    return this->minusInfty(P);
  }
  eval.update(state(), moved.getMove());
  const Player Turn = PlayerTraits<P>::opponent;
  const size_t previous_node_count = nodeCount();

  pv[curDepth()].clear();

  int result;
  // ɽԲǽξˤδؿȻ¹Ѥrecord
  // TODO: class  max_depth ٤
  boost::scoped_ptr<SimpleHashRecord> record_if_unavailable;
  SimpleHashRecord *record 
    = table->allocate(currentHash(), curLimit());
  const bool has_table_record = record;
  if (! record) {
    record_if_unavailable.reset(new SimpleHashRecord());
    record = record_if_unavailable.get(); // memorize ƤϤʤ
  }
  setCurrentRecord(record);
  record->is_king_in_check = EffectUtil::isKingInCheck(Turn, state());
#if 0
  // harmful now
  if (pass_count.loopByBothPass()) {
    return quiesce<Turn>(w);
  }
#endif
  ++node_count;
  int consumption = moved.getLogProb();
  if (record->is_king_in_check)
    consumption = std::min(150, consumption);
  
  // hash μ꤬ʤiid
  if (in_pv 
      && (! record->bestMove().getMove().isNormal())) {
    for (int limit = curLimit() - 200+1; limit > consumption+200;
	 limit -= 200) {
      searchAllMoves<Turn>(moved.getMove(), limit, 
			   record, w);
      if (! record->bestMove().validMove()) {
	Move quiesce_best = record->qrecord.bestMove();
	if (quiesce_best.isNormal())
	  record->setBestMove(SearchMove(MoveLogProb(quiesce_best, 100)));
      }
    }
  }
  if (! in_pv) {
    // null window search
    result = searchAllMoves<Turn>
      (moved.getMove(), consumption, record, 
       Window(w.alpha(P), w.alpha(P)));
  } else {
    // normal search
    result = searchAllMoves<Turn>(moved.getMove(), consumption, 
				  record, w);
  }
  bool extended = false;
  // extension if the alpha value is updated
  if (eval::betterThan(P, result, w.alpha(P))
      && ! w.null()) {
    if (! in_pv || consumption > 100) {
      extended = true;
      ext_limit.add(std::max(consumption - 100, 0));
      result = searchAllMoves<Turn>(moved.getMove(), std::min(100, consumption), record, w);
    }
  }
  ext.add(extended);

  if (has_table_record)
    record->addNodeCount(nodeCount() - previous_node_count);
  return result;
}

template <osl::Player Turn>
int osl::search::AlphaBeta2Tree::
searchAllMoves(Move m, int limit_consumption, SimpleHashRecord *record, 
	       Window w)
{
  const Player P = PlayerTraits<Turn>::opponent;
  recorder.tryMove(MoveLogProb(m, limit_consumption),
		   w.alpha(P), curLimit());
  subLimit(limit_consumption);

  const int result = searchAllMoves<Turn>(record, w);

  addLimit(limit_consumption);
  recorder.recordValue(MoveLogProb(m, limit_consumption),
		       result,eval::betterThan(P, result, w.alpha(P)),
		       curLimit());
  return result;
}

template <osl::Player P>
void osl::search::AlphaBeta2Tree::
testThreatmate(SimpleHashRecord *record, bool in_pv)
{
  if ((! record->is_king_in_check)
      && (! (record && record->threatmate().isThreatmate(P)))
      && tryThreatmate())
  {
    int threatmate_limit = 0;

    threatmate_limit = 4500-curDepth()*1000;
    threatmate_limit = std::max(threatmate_limit, 100);
    if (curLimit() >= table->minimumRecordLimit())
    {
      threatmate_limit = record->qrecord.threatmateNodesLeft(threatmate_limit);
    }
    else
    {
      threatmate_limit /= 2;		// for safety against multiple calls
    }

    if (! in_pv)
      threatmate_limit /= 2;
    if (root_limit < 800)
      threatmate_limit /= 2;

    Move threatmate_move;
    recorder.gotoCheckmateSearch(state(), threatmate_limit);
#ifdef CHECKMATE_COUNT
    size_t count = checkmateSearcher().totalNodeCount();
#endif
    bool threatmate
      = isThreatmateState<P>(threatmate_limit, threatmate_move, 
			     record->qrecord.threatmate_oracle);
#ifdef CHECKMATE_COUNT
    count_threatmate += checkmateSearcher().totalNodeCount() - count;
#endif
    recorder.backFromCheckmateSearch();
    if (!threatmate && threatmate_limit == 0
	&& record->qrecord.threatmateNodesLeft(2)) {
      threatmate = isThreatmateStateShort<P>(2, threatmate_move);
    }
    if (threatmate)
    {
      record->threatmate().setThreatmate(P, threatmate_move);
    }
  }
}

template <osl::Player P>
bool osl::search::AlphaBeta2Tree::
tryCheckmate(SimpleHashRecord *record, bool in_pv, Move& checkmate_move)
{
  int checkmate_limit = 1;
  if (! in_pv) {
    const SimpleHashRecord *parent = lastRecord(1);
    if (! (record->threatmate().mayHaveCheckmate(alt(P))
	   || (parent && parent->threatmate().maybeThreatmate(alt(P)))))
      return false;
    // simulation only
  }
  if (in_pv && root_limit >= 500) {
    int depth = curDepth();
    if (root_limit >= 700 && depth <= 3) {
      if (/*record_memorize &&*/ (depth <= 1))
	checkmate_limit = checkmate::limitToCheckCount(3500);
      else if (/* record_memorize &&*/ (depth == 2))
	checkmate_limit = 1000;
      else if (/* record_memorize &&*/ (depth == 3))
	checkmate_limit = 200;
    }
    else if (((root_limit - curLimit()) <= 500) || (depth <= 5))
    {
      assert(static_cast<unsigned int>(curLimit()) < 4096);
      checkmate_limit = checkmate::limitToCheckCount(std::max(0,curLimit()-200));
    }
    const SimpleHashRecord *parent = lastRecord(1);
    if (record->threatmate().mayHaveCheckmate(alt(P))
	|| (parent && parent->threatmate().maybeThreatmate(alt(P))))
      checkmate_limit += std::max(100, checkmate_limit);
    if (root_limit < 800)
      checkmate_limit /= 2;
  }
  if (curLimit() >= table->minimumRecordLimit())
  {
    // ;ϤٿȤƤ
    checkmate_limit = record->qrecord.checkmateNodesLeft(checkmate_limit);
    if (checkmate_limit <= 0)
      return false;
  }
  else
  {
    checkmate_limit /= 2;		// for safety against multiple calls
  }
  
  // ͤޤȤͤ 
#ifdef CHECKMATE_COUNT
  size_t count = checkmateSearcher().totalNodeCount();
#endif
  recorder.gotoCheckmateSearch(state(), checkmate_limit);
  const bool win = isWinningState<P>
    (checkmate_limit, checkmate_move, record->qrecord.attack_oracle);
  recorder.backFromCheckmateSearch();
#ifdef CHECKMATE_COUNT
  checkmate_before += checkmateSearcher().totalNodeCount() - count;
#endif
  return win;
}

template <osl::Player P>
bool osl::search::AlphaBeta2Tree::
tryCheckmateAgain(SimpleHashRecord *record, Move& checkmate_move,
		  int node_count, int best_value)
{
  int checkmate_limit = 1;
  if (EvalTraits<P>::betterThan(eval.captureValue(newPtypeO(P,KING)), best_value))
    checkmate_limit = node_count / 2; // ζ̤ɬ or ;ʳalt(P)ξ
  else 
    checkmate_limit = node_count / 8;

  checkmate_limit += 100;
  if (true /*record_memorize*/)
    checkmate_limit = record->qrecord.checkmateNodesLeft(checkmate_limit);

#ifdef CHECKMATE_COUNT
  size_t count = checkmateSearcher().totalNodeCount();
#endif
  recorder.gotoCheckmateSearch(state(), checkmate_limit);
  const bool win = isWinningState<P>
    (checkmate_limit, checkmate_move, record->qrecord.attack_oracle);
  recorder.backFromCheckmateSearch();
#ifdef CHECKMATE_COUNT
  checkmate_after += checkmateSearcher().totalNodeCount() - count;
#endif

  return win;
}

bool osl::search::
AlphaBeta2Tree::tryPass(SimpleHashRecord *record, Player P) const
{
  return ! record->is_king_in_check
    && ! record->threatmate().maybeThreatmate(P);
}

template <osl::Player P>
const osl::MoveLogProb osl::search::
AlphaBeta2Tree::nextMove()
{
  MoveGenerator& generator = makeGenerator();
  SimpleHashRecord *record = lastRecord();
  switch (move_type[curDepth()]) {
  case HASH:
  {
    if (curLimit() < leaf_limit) {
      move_type[curDepth()] = FINISH;
      break;
    }
    move_type[curDepth()] = TACTICAL; 
    const MoveLogProb best_move_in_table = record->bestMove().moveLogProb();
    assert(curLimit() > 0);
    generator.init(curLimit(), record, eval.progress32(), state(), 
		   best_move_in_table.getMove());
    if (best_move_in_table.validMove() && 
	validTableMove(state(), best_move_in_table, curLimit())) {
      return best_move_in_table;
    }
  }
  // fall through
  // TODO: ͤϤǥå᤽
  case TACTICAL:
  {
    MoveLogProb m = generator.nextTacticalMove<P>(*this);
    if (m.validMove())
      return m;
    // fall through
    move_type[curDepth()] = KILLER;
    killers[curDepth()].clear();
    if ((! record->is_king_in_check)
	&& ! record->threatmate().maybeThreatmate(P)
	&& (curLimit() >= 300)) {
      MoveVector killer_moves;	// TODO: Ψ
      getKillerMoves(killer_moves);
      for (MoveVector::const_iterator p=killer_moves.begin(); 
	   p!=killer_moves.end(); ++p) {
	assert(killers[curDepth()].size() < killers[curDepth()].capacity());
	killers[curDepth()].push_back(*p);
      }
      std::reverse(killers[curDepth()].begin(), killers[curDepth()].end());
    }
  }
  case KILLER:
  {
    killer_t& killers = this->killers[curDepth()];
    if (! killers.empty()) {
      Move m = killers[killers.size()-1];
      killers.pop_back();
      return MoveLogProb(m, 300);
    }
    // fall through
    move_type[curDepth()] = PASS;
  }
  case PASS:
    assert(record->is_king_in_check == EffectUtil::isKingInCheck(P, state()));
    move_type[curDepth()] = ALL;
    if (tryPass(record, P)) {
      const int pass_consumption = (curLimit() >= 800) ? 300 : 200;
      return MoveLogProb(Move::PASS(P), pass_consumption);
    }
    // TODO: pass θκ
    // fall through
  case ALL:
  {
    MoveLogProb m = generator.nextMove<P>(*this);
    if (m.validMove())
      return m;
    move_type[curDepth()] = FINISH;
  }
  default:
    assert(move_type[curDepth()] == FINISH);
  }
  return MoveLogProb();
}
  

template <osl::Player P>
int osl::search::AlphaBeta2Tree::
searchAllMoves(SimpleHashRecord *record, Window w)
{
#ifndef NDEBUG
  checkPointSearchAllMoves();
#endif
  assert(P == state().getTurn());
  search_assert(w.isConsistent(), lastMove());
  assert(curLimit() >= 0);

  assert(hasLastRecord(1));
  const SimpleHashRecord *parent = lastRecord(1);
  const int node_count_at_beginning = nodeCount();
  depth_node_count[curDepth()]++;

  move_type[curDepth()] = INITIAL;
  const bool in_pv = ! w.null();

  // ơ֥ˤͤĴ٤
  if (record) {
    if (in_pv) {
      if (record->hasLowerBound(SearchTable::CheckmateSpecialDepth)) {
	int lower_bound = record->lowerBound();
	if (isWinValue(P, lower_bound))
	  return lower_bound;
      }
    }
    else {			// ! in_pv
      int table_value = 0;
      if (record->hasGreaterLowerBound<P>(curLimit(), w.alpha(P), 
					  table_value)) {
	assert(eval::isConsistentValue(table_value));
	w.alpha(P) = table_value + EvalTraits<P>::delta;
	if (EvalTraits<P>::betterThan(table_value, w.beta(P))) {
	  recorder.tableHitLowerBound(P, table_value, w.beta(P), curLimit());
	  return table_value;
	}
      } 
      if (record->hasLesserUpperBound<P>(curLimit(), w.beta(P), table_value)) {
	assert(eval::isConsistentValue(table_value));
	w.beta(P) = table_value - EvalTraits<P>::delta;
	if (EvalTraits<P>::betterThan(w.alpha(P), table_value))
	{
	  recorder.tableHitUpperBound(P, table_value, w.alpha(P), curLimit());
	  return table_value;
	}
      }
    }

    Move checkmate_move=Move::INVALID();
    if ((!record->is_king_in_check)
	&& record->qrecord.checkmateNodesLeft(1)
	&& isWinningStateShort<P>(2, checkmate_move))
    {
      recordWinByCheckmate(P, record, checkmate_move);
      return winByCheckmate(P);
    }    
#ifndef DONT_USE_CHECKMATE
    assert(record);
    // try simulation or simple checkmate search
    int additional_limit = 0;		// simulation only
    if (parent && parent->threatmate().maybeThreatmate(alt(P)))
    {
      additional_limit = std::max(100, parent->qrecord.threatmateNodes()/8);
      additional_limit = record->qrecord.checkmateNodesLeft(additional_limit);
    }
    recorder.gotoCheckmateSearch(state(), additional_limit);
    const bool win = isWinningState<P>(additional_limit, checkmate_move, 
				       record->qrecord.attack_oracle);
    recorder.backFromCheckmateSearch();
    if (win) {      
      assert(checkmate_move.isValid());
      recordWinByCheckmate(P, record, checkmate_move);
      return winByCheckmate(P);
    }
#endif
  }

  search_assert(w.isConsistent(), lastMove());
  const int initial_alpha = w.alpha(P);

#ifndef DONT_USE_CHECKMATE
  // ͤͤ
  testThreatmate<P>(record, in_pv);
#endif
  // õ ThreatmateState 
  record->threatmate().update(P, (parent ? &(parent->threatmate()) : 0), 
			      EffectUtil::isKingInCheck(P, state()));

  MoveLogProb best_move;	// invalidated
  int best_value = minusInfty(P);
  int tried_moves = 0;
  int alpha_update = 0;
  int last_alpha_update = 0;
#if (defined OSL_SMP) && (! defined OSL_SMP_NO_SPLIT_INTERNAL)
  bool already_split = false;
#endif

  // the first move
  MoveLogProb m = nextMove<P>();
  ++tried_moves;
  if (! m.validMove() || m.getLogProb() > curLimit()) {
    goto move_generation_failure;
  }
  int first_move_node;
  {
    const int previous_node_count = nodeCount();

    assert(eval::betterThan(P, w.alpha(P), best_value));
    const int result = alphaBetaSearch<P>(m, w, in_pv);
    if (eval::betterThan(P, result, best_value)) {
      best_value = result;
      best_move = m;
      if (eval::betterThan(P, best_value, w.alpha(P))) {
	w.alpha(P) = result + EvalTraits<P>::delta;;
	assert(best_move.validMove());
	++alpha_update;
	last_alpha_update = 1;
	if (eval::betterThan(P, result, w.beta(P))) {
	  mpn_cut.add(tried_moves);
	  if (move_type[curDepth()] >= ALL)
	    setKillerMove(best_move.getMove());
	  assert(best_move.validMove());
	  assert(! isWinValue(alt(P), best_value));
	  goto register_table;
	} else {
	  if (in_pv) 
	    makePV(m.getMove());
	}
      }
    }
    first_move_node = nodeCount() - previous_node_count;
  }
  testStop();

#ifndef DONT_USE_CHECKMATE
  if (in_pv)
  {
    Move checkmate_move;
    if (tryCheckmate<P>(record, in_pv, checkmate_move)) {
      assert(checkmate_move.isValid());
      best_value= winByCheckmate(P);
      recordWinByCheckmate(P, record, checkmate_move);
      return best_value;
    }
  }
#endif
  search_assert(w.isConsistent(), lastMove());
  if (curLimit() < leaf_limit)
    goto move_generation_failure;
  while (true) {
#if (defined OSL_SMP) && (! defined OSL_SMP_NO_SPLIT_INTERNAL)
    const bool prefer_split = curLimit() >= shared->split_min_limit
      && (curLimit() >= 600
	  // || (curLimit() >= 400 && curDepth() < 2)
	  || (first_move_node >= 60000));
    try {
      if (prefer_split && shared->smp_idle) {
	assert(! already_split);
	already_split = true;
	if (examineMovesOther<P>(w, best_move, best_value, tried_moves,
				 alpha_update, last_alpha_update)) {
	  assert(best_move.validMove());
	  assert(best_move.getMove().player() == P);
	  if (move_type[curDepth()] >= ALL)
	    setKillerMove(best_move.getMove());
	  mpn_cut.add(tried_moves);    
	  goto register_table;
	}
	goto all_moves_done;
      }
    }
    catch(SplitFailed&) {
      already_split = false;
      // fall through
    }
#endif
    MoveLogProb m = nextMove<P>(); 
    if (! m.validMove())
      break;
    ++tried_moves;

    assert(eval::betterThan(P, w.alpha(P), best_value));
    const int result = alphaBetaSearch<P>(m, w, in_pv && ! best_move.validMove());
    if (eval::betterThan(P, result, best_value)) {
      best_value = result;
      best_move = m;
      if (eval::betterThan(P, best_value, w.alpha(P))) {
	w.alpha(P) = result + EvalTraits<P>::delta;;
	assert(best_move.validMove());
	++alpha_update;
	last_alpha_update = tried_moves;
	if (eval::betterThan(P, result, w.beta(P))) {
	  assert(best_move.validMove());
	  if (move_type[curDepth()] >= ALL)
	    setKillerMove(best_move.getMove());
	  mpn_cut.add(tried_moves);
	  goto register_table;
	} else {
	  if (in_pv)
	    makePV(m.getMove());
	}
      }
    }
  }
#if (defined OSL_SMP) && (! defined OSL_SMP_NO_SPLIT_INTERNAL)
all_moves_done:
#endif
  if (tried_moves == 1 && tryPass(record, P)) {
    // goto quiescence search if tried move is only null move
    goto move_generation_failure;
  }
  mpn.add(tried_moves);
  if (alpha_update) {
    this->alpha_update.add(alpha_update);
    this->last_alpha_update.add(last_alpha_update);
  }
register_table:
  assert(best_value == minusInfty(P) || best_move.validMove());
  assert(eval::isConsistentValue(best_value));
  if (record && record->threatmate().maybeThreatmate(P)) {
    // ⤦١ͤޤȤͤ
#ifndef DONT_USE_CHECKMATE
    Move checkmate_move=Move::INVALID();
    if (tryCheckmateAgain<P>(record, checkmate_move, 
			     nodeCount() - node_count_at_beginning,
			     best_value)) {
      assert(checkmate_move.isValid());
      best_value= winByCheckmate(P);
      recordWinByCheckmate(P, record, checkmate_move);
      return best_value;
    }
#endif
  }

  if (isWinValue(alt(P), best_value))
  {
    // TODO: ľ꤬(䤱)Ϣ³ä
    // ɬǤϤʤƵͰɤǽϤ
    best_value = brinkmatePenalty(P, curLimit()) + eval.value();

    // (οǤ)==
    record->setAbsoluteValue(best_move, best_value, curLimit());
    return best_value;
  }
  else if (EvalTraits<P>::betterThan(w.alpha(P), initial_alpha)) {
    if (best_move.validMove()) {
      best_move.setLogProb(RealizationProbability::TableMove);
      assert(best_value % 2 == 0);
      record->setLowerBound(P, curLimit(), 
			    SearchMove(best_move), best_value);
    }
  }
  if (EvalTraits<P>::betterThan(w.beta(P), best_value)) {
    if (best_move.validMove()) {
      best_move.setLogProb(RealizationProbability::TableMove);
      record->setUpperBound(P, curLimit(), 
			    SearchMove(best_move), best_value);
    }
  }
  return best_value;
move_generation_failure:
  // Ǥʤä
  // limit  200 ̤äȡʾä move generator  limit ʲμǤʤä礳
  best_value = quiesce<P>(w);
  if (record)
  {
    if (EvalTraits<P>::betterThan(best_value, initial_alpha)) {
      if (EvalTraits<P>::betterThan(w.beta(P), best_value)) {
	record->setAbsoluteValue(MoveLogProb(), best_value, curLimit());
      } else {
	recordLowerBound(P, record, curLimit(), SearchMove(), best_value);
      }
    }
    else 
    {
      assert(EvalTraits<P>::betterThan(w.beta(P), best_value));
      recordUpperBound(P, record, curLimit(), SearchMove(), best_value);
    }
  }
  assert(eval::isConsistentValue(best_value));
  return best_value;
}

template <osl::Player P>
int osl::search::AlphaBeta2Tree::
quiesce(Window w)
{
  testStop();
  initPV();
  
  typedef QuiescenceSearch2<eval::ProgressEval> qsearcher_t;
  qsearcher_t qs(*this, *table);
  Move last_move = lastMove();
  if (last_move.isInvalid())
    last_move = Move::PASS(alt(P));
  assert(w.alpha(P) % 2);
  assert(w.beta(P) % 2);
#ifdef CHECKMATE_COUNT
  size_t count = checkmateSearcher().totalNodeCount();
#endif
  const int result = qs.template search<P>(w.alpha(P), w.beta(P), eval, last_move, 4);
  node_count += qs.nodeCount();
  recorder.addQuiescenceCount(qs.nodeCount());
#ifdef CHECKMATE_COUNT
  quiesce_checkmate += checkmateSearcher().totalNodeCount() - count;
#endif

  assert(result % 2 == 0);
  return result;
}

template <osl::Player P>
int osl::search::AlphaBeta2Tree::
quiesceExp(Window w)
{
  testStop();

  // TODO: Żõε;
  // TODO: futility margin
  SimpleHashRecord *record = lastRecord();
  assert(record);
  Move best_move = record->qrecord.bestMove();
  const int qdepth = 4;
  const int previous_node_count = nodeCount();

  int result = // eval.value();
    quiesceRoot<P>(w, qdepth, best_move, record->threatmate());
  
  const size_t qnode = nodeCount() - previous_node_count;
  recorder.addQuiescenceCount(qnode);
  record->qrecord.setLowerBound(qdepth, result, best_move);
  return result;
}

template <osl::Player P>
struct osl::search::AlphaBeta2Tree::NextQMove
{
  AlphaBeta2Tree *searcher;
  Window window;
  const int depth;
  int *result;
  DualThreatmateState threatmate;
  NextQMove(AlphaBeta2Tree *s, Window w, int d, int *r,
	    DualThreatmateState t)
    : searcher(s), window(w), depth(d), result(r), threatmate(t) {
  }
  void operator()(Position /*last_to*/) {
    searcher->eval.update(searcher->state(), searcher->lastMove());
    *result = 
      searcher->quiesce<P>(window, depth, threatmate);
  }
};

template <osl::Player P>
bool osl::search::AlphaBeta2Tree::
quiesceWithMove(Move move, Window w, int depth_left, Move& best_move, int& best_value, 
		const DualThreatmateState& threatmate)
{
  // TODO: futility margin
  int result;
  typedef NextQMove<PlayerTraits<P>::opponent> next_t;
  next_t helper(this, w, depth_left, &result, threatmate);

  const HashKey new_hash = currentHash().newHashWithMove(move);
  const eval_t old_eval = eval;
  doUndoMoveOrPass<P,next_t>(new_hash, move, helper);
  eval = old_eval;

  if (eval::betterThan(P, result, best_value)) {
    best_value = result;
    best_move = move;
    if (eval::betterThan(P, best_value, w.alpha(P))) {
      w.alpha(P) = result + EvalTraits<P>::delta;
      makePV(best_move);
      if (eval::betterThan(P, result, w.beta(P))) {
	return true;
      }
    }
  }
  return false;
}

template <osl::Player P>
int osl::search::AlphaBeta2Tree::
quiesceRoot(Window w, int depth_left, Move& best_move, DualThreatmateState threatmate)
{
  assert(! EffectUtil::isKingInCheck(alt(P), state()));

  initPV();
  depth_node_count_quiesce[curDepth()]++;
  ++node_count;

  SimpleHashRecord& record = *lastRecord();
  assert(record.is_king_in_check == EffectUtil::isKingInCheck(P, state()));
  assert(depth_left > 0);

  int best_value = minusInfty(P);
  MoveGenerator& generator = makeGenerator();
  generator.init(200, &record, eval.progress32(), state(), 
		 best_move, true);
  int tried_moves = 0, tried_pass = 0;
  // previous 
  if (best_move.isNormal()) {
    ++tried_moves;
    if (quiesceWithMove<P>(best_move, w, depth_left-1, best_move, best_value,
			   threatmate))
      goto finish;
  }

  // stand pat
  if (! record.is_king_in_check 
      && ! threatmate.maybeThreatmate(P)
      && EvalTraits<P>::betterThan(eval.value(), w.alpha(P))) {
    best_move = Move::PASS(P);
    return eval.value();
  } else if (threatmate.maybeThreatmate(P)) {
    const int value = eval.value() + threatmatePenalty(P);
    best_value = EvalTraits<P>::max(best_value, value);
    if (EvalTraits<P>::betterThan(best_value, w.beta(P)))
      goto finish;
  }

  // captures or king escape
  for (MoveLogProb m = generator.nextTacticalMove<P>(*this); 
       m.validMove(); m = generator.nextTacticalMove<P>(*this)) {
    ++tried_moves;
    if (quiesceWithMove<P>(m.getMove(), w, depth_left-1, best_move, best_value,
			   threatmate))
      goto finish;
  } 

  // pawn drop foul?
  if (record.is_king_in_check) {
    if (tried_moves == 0) {
      if (lastMove().ptype() == PAWN && lastMove().isDrop())
	return winByFoul(P);
      return winByCheckmate(alt(P));
    }
    goto finish;
  }
  
  // pass
  if (! tried_pass) {
    if (quiesceWithMove<P>(Move::PASS(P), w, depth_left-1, best_move, best_value,
			   threatmate))
      goto finish;
  }
finish:
  return best_value;
}

template <osl::Player P>
int osl::search::AlphaBeta2Tree::
quiesce(Window w, int depth_left, DualThreatmateState parent_threatmate)
{
  if (EffectUtil::isKingInCheck(alt(P), state())) {
    return this->minusInfty(alt(P));
  }

  initPV();
  depth_node_count_quiesce[curDepth()]++;
  ++node_count;

  SimpleHashRecord record;
  record.is_king_in_check = EffectUtil::isKingInCheck(P, state());

  DualThreatmateState threatmate;
  threatmate.update(P, &parent_threatmate, record.is_king_in_check);

  int best_value = minusInfty(P);
  // TODO: ̤βtakebackĹ
  if (depth_left <= 0) {
    if (record.is_king_in_check) {
      if (lastMove().capturePtype() != PTYPE_EMPTY)
	depth_left +=2;
      else
	depth_left = 0;
    }
    else if (threatmate.maybeThreatmate(P)) {
      if (threatmate.mayHaveCheckmate(alt(P))) {
	Move checkmate_move;
	bool win = isWinningState<P>(10, checkmate_move, record.qrecord.attack_oracle);
	if (win)
	  return winByCheckmate(P);
      }
      return best_value;
    }
    else {
      if (threatmate.mayHaveCheckmate(alt(P))
	  || ImmediateCheckmate::hasCheckmateMove<P>(state()))
	return eval.value() + threatmatePenalty(alt(P));
      if (ImmediateCheckmate::hasCheckmateMove<PlayerTraits<P>::opponent>(state()))
	return eval.value() + threatmatePenalty(P);
      return eval.value();
    }
  }

  if (! record.is_king_in_check) {
    if (ImmediateCheckmate::hasCheckmateMove<P>(state())) {
      return winByCheckmate(P);
    }
  }
  MoveGenerator& generator = makeGenerator();
  generator.init(200, &record, eval.progress32(), state(), 
		 Move(), true);
  int tried_moves = 0, tried_pass = 0;
  Move best_move;
  // stand pat
  if (! record.is_king_in_check 
      && ! threatmate.maybeThreatmate(P)
      && EvalTraits<P>::betterThan(eval.value(), w.alpha(P))) {
    best_move = Move::PASS(P);
    return eval.value();
  } 

  // captures or king escape
  for (MoveLogProb m = generator.nextTacticalMove<P>(*this); 
       m.validMove(); m = generator.nextTacticalMove<P>(*this)) {
    ++tried_moves;
    if (quiesceWithMove<P>(m.getMove(), w, depth_left-1, best_move, best_value,
			   threatmate))
      goto finish;
  } 

  // pawn drop foul?
  if (record.is_king_in_check) {
    if (tried_moves == 0) {
      if (lastMove().ptype() == PAWN && lastMove().isDrop()) 
	return winByFoul(P);
      return winByCheckmate(alt(P));
    }
    goto finish;
  }
  
  // pass
  if (! tried_pass) {
    if (quiesceWithMove<P>(Move::PASS(P), w, depth_left-1, best_move, best_value,
			   threatmate))
      goto finish;
  }
finish:
  return best_value;
}

void osl::search::AlphaBeta2Tree::updateCheckmateCount()
{
#ifdef OSL_SMP
  recorder.setCheckmateCount(shared->checkmateCount());
#else
  recorder.setCheckmateCount
    (checkmateSearcher().totalNodeCount());
#endif
}

void osl::search::AlphaBeta2Tree::
showPV(std::ostream& os, int result, Move m) const
{
#ifdef OSL_SMP
  boost::mutex::scoped_lock lk(shared->lock_io);
#endif
  assert(m.isNormal()); // do not pass at root
  MoveVector moves;
  size_t quiesce_start = (size_t)-1;
  moves.push_back(m);
  HashKey key = currentHash().newHashWithMove(m);
  table->getPV(key, moves, &quiesce_start);
#ifdef SHOW_ALSO_OLD_PV
  os << "! " << std::setfill(' ') << std::setw(5) 
     << result*100/eval.captureValue(newPtypeO(WHITE,PAWN))*2 << " ";
  for (size_t i=0; i<moves.size(); ++i) {
    if (i == quiesce_start)
      os << " | ";
    os << record::csa::show(moves[i]);
  }
  os << std::endl;
#endif
  if (root_ignore_moves)
    os << "[" << root_ignore_moves->size() << "] ";
  os << "  " << std::setfill(' ') << std::setw(5) 
     << result*100/eval.captureValue(newPtypeO(WHITE,PAWN))*2 << " ";
  for (size_t i=0; i<pv[0].size(); ++i) {
    os << record::csa::show(pv[0][i]);
  }
  os << std::endl;
}

template <osl::Player P>
struct osl::search::AlphaBeta2Tree::NextMove
{
  AlphaBeta2Tree *searcher;
  const MoveLogProb& moved;
  Window window;
  int *result;
  bool in_pv;
  NextMove(AlphaBeta2Tree *s, const MoveLogProb& md, Window w, int *r,
	   bool p)
    : searcher(s), moved(md), window(w), result(r), in_pv(p) {
    assert(P == md.getMove().player());
  }
  void operator()(Position /*last_to*/) {
#ifndef NDEBUG
    const int cur_limit = searcher->curLimit();
#endif
    *result = 
      searcher->alphaBetaSearchAfterMove<P>(moved, window, in_pv);
    assert(cur_limit == searcher->curLimit() || searcher->SearchState2Core::abort());
  }
};

template <osl::Player P>
int osl::search::AlphaBeta2Tree::
alphaBetaSearch(const MoveLogProb& search_move, Window w, bool in_pv)
{
  assert(w.alpha(P) % 2);
  assert(w.beta(P) % 2);
  const Move move = search_move.getMove();
  assert(P == move.player());
  assert(P == state().getTurn());
  assert(eval::notLessThan(P, w.beta(P), w.alpha(P)));

  if ((! timerAvailable()) && (! timer.isInvalid()) && (! timer.timeLeft()))
    throw misc::NoMoreTime();
  // TODO: more efficient way to find foul
  if (! move.isPass()
      && (move_classifier::MoveAdaptor<move_classifier::PawnDropCheckmate<P> >
	  ::isMember(state(), move))) {
    return winByFoul(alt(P));
  }

  const HashKey new_hash = currentHash().newHashWithMove(move);
  assert(P == move.player());

  if (move.isPass())
    pass_count.inc(P);

  // ǧ
  if (! pass_count.loopByBothPass()) {
    const Sennichite next_sennichite
      = repetition_counter.isAlmostSennichite(new_hash);
    if (next_sennichite.isDraw())
      return drawValue();
    if (next_sennichite.hasWinner())
      return winByFoul(next_sennichite.winner());
    assert(next_sennichite.isNormal());
  }

  if (! move.isPass()) {
    // ͥ۴طǧ
    const DominanceCheck::Result has_dominance
      = DominanceCheck::detect(repetition_counter.history(), new_hash);
    if (has_dominance == DominanceCheck::LOSE)
      return winByLoop(alt(P));
    if (has_dominance == DominanceCheck::WIN)
      return winByLoop(P);
    // Ϣ³Τ
    if (move.capturePtype() == PTYPE_EMPTY) {
      const int sacrifice_count = countSacrificeCheck2(curDepth());
      if (sacrifice_count == 2) {
	// 3ܤϻؤʤ
	const Position to = move.to();
	int offence = state().countEffect(P, to) + (move.isDrop() ? 1 : 0);
	const int deffense = state().hasEffectBy(alt(P), to); // max1
	if (offence <= deffense)
	  offence += AdditionalEffect::count2(state(), to, P);
	if (offence <= deffense) {
	  return winByLoop(alt(P));
	}
      }
    }
  }

  // õ
  int result;
  NextMove<P> helper(this, search_move, w, &result, in_pv);

  recorder.addNodeCount();  
  const eval_t old_eval = eval;
  doUndoMoveOrPass<P,NextMove<P> >
    (new_hash, move, helper);
  eval = old_eval;

  if (move.isPass())
    pass_count.dec(P);

  return result;
}

template <osl::Player P>
void osl::search::AlphaBeta2Tree::
examineMovesRoot(const MoveLogProbVector& moves, size_t i, Window window,
		 MoveLogProb& best_move, int& best_value)
{
  for (;i<moves.size(); ++i) {
    testStop();
    const MoveLogProb& m = moves[i];
    const int result = alphaBetaSearch<P>(m, window, false);
    if (eval::betterThan(P, result, best_value)) {
      window.alpha(P) = result + EvalTraits<P>::delta;
      best_move = m;
      best_value = result;
      makePV(m.getMove());
      if (table->isVerbose()) {
	showPV(std::cerr, result, m.getMove());
      }
      if (eval::betterThan(P, result, window.beta(P))) {
	assert(! isWinValue(alt(P), result));
	break;
      }
    }
  }
}

/* ------------------------------------------------------------------------- */

osl::search::AlphaBeta2::
AlphaBeta2(const HashEffectState& s, checkmate_t& c, 
	   SimpleHashTable *t, CountRecorder& r)
  : AlphaBeta2Tree(s, c, t, r)
{
}

osl::search::AlphaBeta2::
AlphaBeta2(const AlphaBeta2& src)
  : AlphaBeta2Tree(src)
{
  // copy outside of search
}

osl::search::AlphaBeta2::
~AlphaBeta2()
{
}

osl::Move osl::search::AlphaBeta2::
computeBestMoveIteratively(int limit, const int step, 
			   const int initial_limit, int node_limit,
			   const misc::RealTime& initial_stop_shedule)
{
  setStopSchedule(initial_stop_shedule);

  const bool use_signal_timer = hasSchedule();
  if (use_signal_timer)
  {
    this->timer = stopSchedule();
    setTimeOut(timer.getTimeLeftInSeconds());
  }
  else
  {
    this->timer = misc::RealTime(6000);
  }

  recorder.resetNodeCount();

  double last_iteration_consumed = 0;
  double total_consumed = 0;
  int limit_iterative = initial_limit;
  Move last_best_move = Move::INVALID();

#ifdef OSL_SMP
#  ifdef SPLIT_STAT
  shared->parallel_splits = 0;
  shared->cancelled_splits.setValue(0);
  shared->parallel_abort.setValue(0);
#  endif
#endif
  try
  {
    MoveLogProb search_move;
    alphaBetaSearchRoot(search_move, 0);
    if (table->verboseLevel() > 1)
      std::cerr << "=> quiesce "
		<< record::csa::show(search_move.getMove()) << "\n";
    while (limit_iterative < limit)
    {
      this->timer = stopSchedule();
      if (table->verboseLevel() > 1)
	std::cerr << "=> iteration " << limit_iterative 
		  << " " << timer.getTimeLeftInDouble() << " left ("
		  << last_iteration_consumed << ", " << total_consumed << ")"  << "\n";
      recorder.startSearch(limit_iterative);
      const int previous_node_count = nodeCount();
      alphaBetaSearchRoot(search_move, limit_iterative);

      last_best_move = search_move.getMove();
      last_iteration_consumed = timer.getConsumedInDouble() - total_consumed;
      total_consumed += last_iteration_consumed;

      updateCheckmateCount();
      if (table->verboseLevel() > 1) {
	std::cerr << "<= " 
		  << record::csa::show(search_move.getMove());
	std::cerr << std::setprecision(3) << "  mpn " << mpn.getAverage() 
		  << " cut " << mpn_cut.getAverage()
		  << " alpha " << alpha_update.getAverage()
		  << " last " << last_alpha_update.getAverage()
		  << " ext " << ext.getAverage()
		  << " ext_limit " << ext_limit.getAverage();
#ifdef OSL_SMP
#  ifdef SPLIT_STAT
	std::cerr << " split " << shared->parallel_splits << " cancel " << shared->cancelled_splits.value() 
		  << " abort " << shared->parallel_abort.value();
#  endif
#endif
	std::cerr << "\n";
      }
      bool time_over = false;
      if (hasSchedule()) {
	const double current_time_left = stopSchedule().getTimeLeftInDouble();
	const double coef = nextIterationCoefficient();
	if (current_time_left 
	    < last_iteration_consumed * coef)
	  time_over = true;
	if (! time_over) {
	  SimpleHashRecord *record
	    = table->find(currentHash());
	  if (record) {
	    record->addNodeCount(nodeCount() - previous_node_count);
	  }
	}
      }
      const bool checkmate_over = recorder.checkmateCount() > CHECKMATE_DEFAULT_TOTAL_NODE_LIMIT;
      bool node_limit_over = (recorder.nodeCount() *4 > node_limit)
	|| checkmate_over;
      recorder.finishSearch(search_move.getMove(), 
			    total_consumed,
			    (time_over || node_limit_over) && table->verboseLevel() > 1);
      if (time_over || node_limit_over) {
	if (table->isVerbose())
	  std::cerr << "iteration stop at " << limit_iterative << " by "
		    << (time_over ? "time" : (checkmate_over ? "checkmate count" : "node count")) << "\n";
	goto finish;
      }
      testStop();
      limit_iterative += step;
    }
    if (table->verboseLevel() > 1)
      std::cerr << "=> final iteration " << limit_iterative 
		<< " " << timer.getTimeLeftInDouble() << " left ("
		<< last_iteration_consumed << ", " << total_consumed << ")"  << "\n";
    while (true) {
      recorder.startSearch(limit);
      try {
	alphaBetaSearchRoot(search_move, limit);
      } catch (...) {
	updateCheckmateCount();
	last_iteration_consumed = timer.getConsumedInDouble() - total_consumed;
	total_consumed += last_iteration_consumed;
	recorder.finishSearch(search_move.getMove(), total_consumed,
			      table->verboseLevel() > 1);
	throw;
      }
      last_iteration_consumed = timer.getConsumedInDouble() - total_consumed;
      total_consumed += last_iteration_consumed;
      updateCheckmateCount();
      recorder.finishSearch(search_move.getMove(), total_consumed,
			    table->verboseLevel() > 1);
      last_best_move = search_move.getMove();

      if (last_best_move.isNormal())
	break;
      testStop();

      // ۤäƤλ
      if (limit >= 2000)
	break;

      limit += 200;
      if (table->isVerbose())
	std::cerr << "  extend limit to " << limit << " before resign\n";
    }
  }
  catch (std::exception& e)
  {
    std::cerr << "std exception " << e.what() << "\n";
  }
  catch (...)
  {
    std::cerr << "unknown exception\n";
#ifndef NDEBUG
    throw;
#endif
  }
finish:
  if (use_signal_timer)
    resetTimeOut();
  if (table->verboseLevel() > 1) {
    std::cerr << "<= " << record::csa::show(last_best_move);
    std::cerr << "  mpn " << mpn.getAverage()
	      << " cut " << mpn_cut.getAverage()
	      << " alpha " << alpha_update.getAverage()
	      << " last " << last_alpha_update.getAverage()
	      << " ext " << ext.getAverage()
	      << " ext_limit " << ext_limit.getAverage();
#ifdef OSL_SMP
#  ifdef SPLIT_STAT
    std::cerr << " split " << shared->parallel_splits << " cancel " << shared->cancelled_splits.value() 
	      << " abort " << shared->parallel_abort.value();
#  endif
#endif
    std::cerr << "\n";
  }
  return last_best_move;
}

template <osl::Player P>
int osl::search::AlphaBeta2::
alphaBetaSearchRoot(Window window, MoveLogProb& best_move, int limit)
{
  assert(P == state().getTurn());
  assert(window.alpha(P) % 2);
  assert(window.beta(P) % 2);
  setRoot(limit);
  assert(curDepth() == 0);

  // ޤơ֥򸣤
  SimpleHashRecord *record_in_table
    = table->allocate(currentHash(), limit);
  SimpleHashRecord *record = record_in_table;
  boost::scoped_ptr<SimpleHashRecord> record_if_not_allocated;
  if (! record)
  {
    record_if_not_allocated.reset(new SimpleHashRecord());
    record = record_if_not_allocated.get();
  }
  assert(record);
  setRootRecord(record);
  assert(rootRecord() == record);
  assert(hasLastRecord() && lastRecord() == record);

  if (limit == 0) {
    int result = quiesce<P>(fullWindow(P));
    best_move = MoveLogProb(record->qrecord.bestMove(), 100);
    if (root_ignore_moves
	&& root_ignore_moves->isMember(best_move.getMove()))
      best_move = MoveLogProb();
    return result;
  }
  if (record_in_table) {
    int table_value = 0;
    const MoveLogProb m = record_in_table->bestMove().moveLogProb();
    if (! m.getMove().isNormal())
      record_in_table->resetValue();
    else if (record->hasGreaterLowerBound<P>(curLimit(), window.beta(P), 
					     table_value)) {
      if (! root_ignore_moves 
	  || ! root_ignore_moves->isMember(m.getMove())) {
	best_move = m;
	return table_value;
      }
    }
  }

  // gather all moves
  MoveLogProbVector moves;
  MoveGenerator& generator = makeGenerator();
  const MoveLogProb last_best_move = record->bestMove().moveLogProb();
  {
    MoveLogProbVector raw_moves;
    record->is_king_in_check = EffectUtil::isKingInCheck(P, state());
    assert(curLimit() > 0);
    generator.init(curLimit(), record, eval.progress32(), state(), last_best_move.getMove());
    if (last_best_move.getMove().isNormal())
      raw_moves.push_back(last_best_move);
    generator.generateAll<P>(*this, raw_moves);

    // clean up losing moves
    for (size_t i=0; i<raw_moves.size(); ++i) {
      const HashKey key = currentHash().newHashWithMove(raw_moves[i].getMove());
      const SimpleHashRecord *record = table->find(key);
      if (record) {
	if (record->hasLowerBound(SearchTable::CheckmateSpecialDepth)
	    && isWinValue(alt(P), record->lowerBound()))
	  continue;
      }
      if (root_ignore_moves && root_ignore_moves->isMember(raw_moves[i].getMove()))
	continue;
      moves.push_back(raw_moves[i]);
    }
  }

  if (moves.size() == 1) {
    best_move = moves[0];
    return (window.alpha(P)+window.beta(P))/2;	// XXX
  }

#ifndef DONT_USE_CHECKMATE
  // ;ƤǤߤ root ǤƤǤʤ
  int checkmate_node = 0;
  if (root_ignore_moves == 0) {
    int checkmate_max = 30000*std::max(limit - 300, 0)/100;
    if (limit >= 1000)
      checkmate_max = std::min(400000, 60000*(limit - 800)/100);
    checkmate_node = record->qrecord.checkmateNodesLeft(checkmate_max);
#ifdef CHECKMATE_COUNT
    std::cerr << "limit " << limit << " checkmate " << checkmate_node << "\n";
#endif
  }
  if (checkmate_node > 0)
  {
    const bool my_king_in_check
      = state().hasEffectBy(alt(P),state().getKingPosition(P));
    if (my_king_in_check)
    {
      // ꤫鲦꤬äƤ
      recorder.gotoCheckmateSearch(state(), checkmate_node);
      const bool lose = isLosingState<P>(checkmate_node/8);
      recorder.backFromCheckmateSearch();
      if (lose)
      {
	best_move = MoveLogProb(Move::INVALID(),100);
	recordLoseByCheckmate(P, record);
	return winByCheckmate(alt(P));
      }
    }
    // ͤޤʤСͤޤȤͤ 
    {
      Move checkmate_move;
#ifdef CHECKMATE_COUNT
      size_t count = checkmateSearcher().totalNodeCount();
#endif
      recorder.gotoCheckmateSearch(state(), checkmate_node);
      const bool win = isWinningState<P>
	(checkmate_node, checkmate_move, record->qrecord.attack_oracle);
      recorder.backFromCheckmateSearch();
#ifdef CHECKMATE_COUNT
      root_checkmate += checkmateSearcher().totalNodeCount() - count;
#endif
      if (win)
      {
	best_move = MoveLogProb(checkmate_move,100);
	recordWinByCheckmate(P, record, checkmate_move);
	return winByCheckmate(P);
      }
    }
    // ͤͤ
    if ((! my_king_in_check)
	&& (! (record->threatmate().isThreatmate(P))))
    {
      Move threatmate_move;
#ifdef CHECKMATE_COUNT
      size_t count = checkmateSearcher().totalNodeCount();
#endif
      recorder.gotoCheckmateSearch(state(), checkmate_node);
      const bool threatmate 
	= isThreatmateState<P>
	(checkmate_node, threatmate_move, record->qrecord.threatmate_oracle);
#ifdef CHECKMATE_COUNT
      root_checkmate += checkmateSearcher().totalNodeCount() - count;
#endif
      recorder.backFromCheckmateSearch();
      if (threatmate)
      {
	if (record)
	  record->threatmate().setThreatmate(P, threatmate_move);
	if (table->verboseLevel() > 1)
	  std::cerr << "  root threatmate " << threatmate_move << "\n";
      }
    }

  }
#endif

  const int ValueNone = window.alpha(P) - EvalTraits<P>::delta;
  int best_value = ValueNone;
  try {
    // first move
    size_t i=0;
    for (;i<moves.size() && best_value == ValueNone; ++i) {
      const MoveLogProb& m = moves[i];
      const int result = alphaBetaSearch<P>(m, window, true);
      if (eval::betterThan(P, result, best_value)) {
	window.alpha(P) = result + EvalTraits<P>::delta;
	best_move = m;
	best_value = result;
	makePV(m.getMove());
	if (table->isVerbose()) {
	  showPV(std::cerr, result, m.getMove());
	}
	if (eval::betterThan(P, result, window.beta(P))) {
	  assert(! isWinValue(alt(P), result));
	}
      }
    }
    if (! eval::betterThan(P, window.alpha(P), window.beta(P))) {
      // others
#if (defined OSL_SMP) && (! defined OSL_SMP_NO_SPLIT_ROOT)
      examineMovesRootPar<P>(moves, i, window, best_move, best_value);
#else
      examineMovesRoot<P>(moves, i, window, best_move, best_value);
#endif
    }
  } catch (misc::NoMoreTime&) {
    if (best_move.validMove()
	&& best_move.getMove() != last_best_move.getMove()) {
      if (table->verboseLevel() > 1)
	std::cerr << "! use better move than the last best move\n";
    }
    else {
#ifdef OSL_SMP
      shared->waitAll();
#endif      
      throw;
    }
  }
  
  assert(best_value % 2 == 0);
  record->setLowerBound(P, curLimit(), SearchMove(best_move),
			best_value);
#ifdef OSL_SMP
  shared->waitAll();
#endif      
  return best_value;
}


void osl::search::AlphaBeta2::setRoot(int limit)
{
  SearchState2::setRoot(limit);
  SimpleHashRecord *record = table->allocate(currentHash(), limit);
  assert(record);
  setRootRecord(record);  
  move_type[curDepth()] = INITIAL;
}

void osl::search::AlphaBeta2::makeMove(Move move)
{
  assert(state().isValidMove(move));
  SearchState2::makeMove(move);
  eval.update(state(), move);

  SimpleHashRecord *record 
    = table->allocate(currentHash(), curLimit());
  assert(record);
  move_type[curDepth()] = INITIAL;
  record->is_king_in_check = EffectUtil::isKingInCheck(state().getTurn(), state());
  setCurrentRecord(record);
}

bool osl::search::AlphaBeta2::
isReasonableMove(Move /*move*/, int /*pawn_sacrifice*/)
{
  return true;
}


void osl::search::AlphaBeta2::
showNodeDepth(std::ostream& os)
{
  int max_depth=0;
  for (int i=MaxDepth-1; i>=0; --i) {
    if (depth_node_count[i] || depth_node_count_quiesce[i]) {
      max_depth = i;
      break;
    }
  }
  int max_count=0;
  for (int i=0; i<=max_depth; i+=2) {
    max_count = std::max(max_count, 
			 depth_node_count[i]+depth_node_count_quiesce[i]);
  }

  int unit = std::max(max_count/79, 100);
  for (int i=0; i<=max_depth; i+=2) {
    os << std::setw(3) << i << " " 
       << std::string(depth_node_count[i]/unit, '*')
       << std::string(depth_node_count_quiesce[i]/unit, '+')
       << std::endl;
  }
#ifdef CHECKMATE_COUNT
  std::cerr << "checkmate root " << root_checkmate << " quiesce " << quiesce_checkmate
	    << "\nnormal before " << checkmate_before 
	    << " after " << checkmate_after << " threatmate " << count_threatmate
	    << "\n";
#endif
}

void osl::search::AlphaBeta2::
clearNodeDepth()
{
  depth_node_count.fill(0);
  depth_node_count_quiesce.fill(0);  
}


namespace osl
{
  namespace search
  {
    template
    void AlphaBeta2Tree::examineMovesRoot<BLACK>(const MoveLogProbVector&, size_t, Window, MoveLogProb&, int&);
    template
    void AlphaBeta2Tree::examineMovesRoot<WHITE>(const MoveLogProbVector&, size_t, Window, MoveLogProb&, int&);
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
