/* moveGenerator.cc
 */
#include "osl/search/moveGenerator.h"
#include "osl/search/searchState2.h"
#include "osl/search/shouldPromoteCut.h"
#include "osl/search/sortCaptureMoves.h"
#include "osl/search/quiescenceGenerator.h"
#include "osl/category/standardMoveGenerator.h"
#include "osl/category/breakThreatmate.h"
#include "osl/category/bigramPosition.h"
#include "osl/category/analyzer/categoryMoveVector.h"
#include "osl/move_generator/capture_.h"
#include "osl/move_generator/escape_.h"
#include "osl/move_generator/promote_.h"
#include "osl/move_generator/addEffect_.h"
#include "osl/move_generator/allMoves.h"
#include "osl/move_generator/attackToPinned.h"
#include "osl/move_action/store.h"
#include "osl/move_classifier/check_.h"
#include "osl/move_classifier/kingOpenMove.h"
#include "osl/effect_util/effectUtil.h"
#include "osl/effect_util/pin.h"
#include "osl/effect_util/neighboring8Direct.h"
#include "osl/effect_util/additionalEffect.h"
#include "osl/rating/featureSet.h"
#include "osl/rating/ratingEnv.h"
#include "osl/eval/pieceEval.h"
#include "osl/neighboring8.h"
#include "osl/ptypeTable.h"
#include "osl/directionTraits.h"
#include "osl/boardTable.h"
#include <iostream>

#ifndef NDEBUG
#  define SAFE_MOVE_ONLY
#endif

static const osl::rating::FeatureSet& feature_set()
{
  static osl::rating::StandardFeatureSet feature_set;
  return feature_set;
}

namespace osl
{
  namespace search
  {
    const CArray2d<MoveGenerator::generator_t, 2, MoveGenerator::FINISH> MoveGenerator::Generators = {{
      {
	0,
	&MoveGenerator::generateKingEscape<BLACK>,
	&MoveGenerator::generateTakeBack<BLACK>,
	&MoveGenerator::generateBreakThreatmate<BLACK>,
	&MoveGenerator::generateCapture<BLACK>,
	0,
	&MoveGenerator::generateTesuji<BLACK>,
	&MoveGenerator::generateAll<BLACK>,
      },
      {
	0,
	&MoveGenerator::generateKingEscape<WHITE>,
	&MoveGenerator::generateTakeBack<WHITE>,
	&MoveGenerator::generateBreakThreatmate<WHITE>,
	&MoveGenerator::generateCapture<WHITE>,
	0,
	&MoveGenerator::generateTesuji<WHITE>,
	&MoveGenerator::generateAll<WHITE>,
      }
    }};
    const CArray<const char *, MoveGenerator::FINISH> MoveGenerator::GeneratorNames = {{
      "INITIAL", "KING_ESCAPE", "TAKEBACK", "BREAK_THREATMATE", "TACTICAL", "SENTINEL", "TESUJI", "ALL", 
    }};
  }
}

/* ------------------------------------------------------------------------- */

osl::search::
MoveMarker::MoveMarker() : cur(1) 
{
  marker.fill(0);
}

void osl::search::
MoveMarker::clear() 
{
  ++cur;
  if (cur == 0) {
    marker.fill(0);
    cur = 1;
  }
}

bool osl::search::
MoveMarker::registerIfNew(const NumEffectState& state, Move m)
{
  value_t& val = marker(toIndex(m), pieceIndex(state, m));
  if (val == cur)
    return false;
  val = cur;
  return true;
}

bool osl::search::
MoveMarker::registered(const NumEffectState& state, Move m) const
{
  return marker(toIndex(m), pieceIndex(state, m)) == cur;
}

/* ------------------------------------------------------------------------- */

osl::search::
MoveGenerator::MoveGenerator() : record(0), progress(16)
{
}

int osl::search::
MoveGenerator::captureValue(Ptype ptype)
{
  if (! isPiece(ptype))
    return 0;
  int result 
    = eval::Ptype_Eval_Table.captureValue(newPtypeO(WHITE, ptype));
  assert(result >= 0);
  return result;
}

void osl::search::
MoveGenerator::init(int l, SimpleHashRecord *r, Progress32 pr,
		    const NumEffectState& state, Move hash_move,
		    bool quiesce) 
{
  assert(r);
  assert(l > 0);
  limit = l;
  record = r;
  cur_state = INITIAL;
  moves.clear();
  bad_capture.clear();
  max_see = cur_index = tried = 0;
  progress = pr;

  marker.clear();
  env.my_pin = effect_util::Pin::make(state, state.getTurn());
  env.op_pin = effect_util::Pin::make(state, alt(state.getTurn()));
  env.sendoffs.clear();
  if (hash_move.isNormal())
    marker.registerMove(state, hash_move);
  in_quiesce = quiesce;
}

void osl::search::
MoveGenerator::dump() const
{
  std::cerr << "generator " << cur_state << " index " << cur_index 
	    << " limit " << limit << " tried " << tried << "\n";
  std::cerr << moves.size() << "\n" << moves;
}

template <osl::Player P>
const osl::MoveLogProb osl::search::
MoveGenerator::nextTacticalMoveWithGeneration(const SearchState2& state) 
{	
  assert(record);
  while (true) {
    assert(cur_index >= moves.size());
    if (cur_state == KING_ESCAPE && record->is_king_in_check) {
      cur_state = FINISH;
      break;
    }
    if (++cur_state >= TACTICAL_FINISH)
      break;
    // generate
    assert(Generators[playerToIndex(P)][cur_state]);
    (this->*Generators[playerToIndex(P)][cur_state])(state);
    if (cur_index < moves.size()) {
      ++tried;
      return moves[cur_index++];
    } 
  }
  return MoveLogProb();
}

template <osl::Player P>
const osl::MoveLogProb osl::search::
MoveGenerator::nextMoveWithGeneration(const SearchState2& state) 
{	
  assert(record);
  while (true) {
    assert(cur_index >= moves.size());
    if (++cur_state >= FINISH)
      break;
    // generate
    assert(Generators[playerToIndex(P)][cur_state]);
    (this->*Generators(playerToIndex(P), cur_state))(state);
    if (cur_index < moves.size()) {
      ++tried;
      return moves[cur_index++];
    } 
  }
  return MoveLogProb();
}

template <osl::Player P>
void osl::search::
MoveGenerator::generateKingEscape(const SearchState2& sstate)
{
  env.history = sstate.history();
  if (! record->is_king_in_check) 
    return;
  using namespace move_action;

  const NumEffectState& state = sstate.state();
  const Piece king = state.getKingPiece<P>();
  assert(state.hasEffectBy(alt(P), king.position()));

  MoveVector src;
  Store store(src);
  move_generator::Escape<P,NumEffectState,Store>::
    generateMoves(state,king,store);

  if (src.size() == 1) {
    moves.push_back(MoveLogProb(src[0], 20));
    return;
  }
  for (MoveVector::const_iterator p=src.begin(); p!=src.end(); ++p) {
    const int prob = std::min(limit, feature_set().logProbKingEscape(state, env, *p));
    assert(prob > 0);
    moves.push_back(MoveLogProb(*p, prob));
  }
  moves.sortByProbability();
}

template <osl::Player P>
void osl::search::
MoveGenerator::generateBreakThreatmate(const SearchState2& sstate)
{
  const NumEffectState& state = sstate.state();
  const category::CategoryEnv env(&state, limit, &sstate.history(), progress,
				  record);
  category::BreakThreatmate::generate(env, moves);
  for (size_t i=0; i<moves.size(); ++i)
    marker.registerMove(state, moves[i].getMove());
}

template <osl::Player P>
void osl::search::
MoveGenerator::generateTakeBack(const SearchState2& sstate)
{
  using namespace move_action;
  const Move last_move = sstate.lastMove();
  if (! last_move.isNormal())
    return;
  const Position last_to = last_move.to();
  
  const NumEffectState& state = sstate.state();
  MoveVector src;
  Store store(src);

  if (in_quiesce)
    quiesceCapture<P>(state, last_to, src);
  else
    move_generator::Capture<P>::generate(state, last_to, store);
  const Position king_position = state.getKingPosition(alt(state.getTurn()));
  effect_util::SendOffPosition::find(state.getTurn(), state, king_position, env.sendoffs);
  assert(moves.empty());
  for (MoveVector::const_iterator p=src.begin(); p!=src.end(); ++p) {
    if (ShouldPromoteCut::canIgnoreMove<P>(*p))
      continue;
    const int prob = feature_set().logProbTakeBack(state, env, *p);
#ifdef OSL_SMP
    if (! p->isDrop() && p->ptype() != KING
	&& env.my_pin.test(state.getPieceOnBoard(p->from()).number())) {
      if (move_classifier::KingOpenMove<P>::isMember(state, p->ptype(), p->from(), p->to()))
	continue;
    }
#endif
    if (prob <= std::min(200, limit) && marker.registerIfNew(state, *p))
      moves.push_back(MoveLogProb(*p, prob));
  }
  moves.sortByProbability();
}

namespace osl
{
  template <Player P>
  static void makeCapture(const NumEffectState& state, int first, int last,
			  MoveVector& out)
  {
    move_action::Store store(out);
    for (int i = first; i < last; i++) {
      const Piece p = state.getPieceOf(i);
      if (p.isOnBoardByOwner<PlayerTraits<P>::opponent>())
	move_generator::Capture<P>::generate(state, p.position(), store);
    }
  }
}

template <osl::Player P>
bool osl::search::
MoveGenerator::addCapture(const NumEffectState& state, const RatingEnv& env, const MoveVector& src)
{
  bool added = false;  
  for (MoveVector::const_iterator p=src.begin(); p!=src.end(); ++p) {
    const Move m = *p;
    if (ShouldPromoteCut::canIgnoreMove<P>(m))
      continue;
#ifdef SAFE_MOVE_ONLY
    if (! m.isDrop() && m.ptype() != KING
	&& env.my_pin.test(state.getPieceOnBoard(m.from()).number())) {
      if (move_classifier::KingOpenMove<P>::isMember(state, m.ptype(), m.from(), m.to()))
	continue;
    }
#endif
    const int prob = feature_set().logProbTakeBack(state, env, *p);
    if (prob <= 200 && marker.registerIfNew(state, *p)) {
      added = true;
      moves.push_back(MoveLogProb(*p, prob));
    }
  }
  return added;
}

template <osl::Player P>
void osl::search::
MoveGenerator::generateCapture(const SearchState2& sstate)
{
  using namespace move_action;

  const NumEffectState& state = sstate.state();
  MoveVector src;
  // lance, bishop, rook
  makeCapture<P>(state, PtypeTraits<LANCE>::indexMin, PtypeTraits<ROOK>::indexLimit,
		 src);		
  // knight, silver, gold
  makeCapture<P>(state, PtypeTraits<KNIGHT>::indexMin, PtypeTraits<GOLD>::indexLimit, src);
  addCapture<P>(state, env, src);
}

template <osl::Player P>
void osl::search::
MoveGenerator::generateTesuji(const SearchState2& sstate)
{
}

template <osl::Player P>
void osl::search::
MoveGenerator::quiesceCapture(const NumEffectState& state, Position to, MoveVector& out)
{
  MoveVector moves;
  move_action::Store store(moves);
  move_generator::Capture<P>::generate(state, to, store);
  int defense = state.countEffect(alt(P), to);
  int target_value = captureValue(state.getPieceOnBoard(to).ptype());
  for (MoveVector::const_iterator p=moves.begin(); p!=moves.end(); ++p) {
    int value = target_value;
    if (defense)
      value -= captureValue(p->oldPtype());
    if (value < 0)
      continue;
    out.push_back(*p);
  }
}

template <osl::Player P>
void osl::search::
MoveGenerator::generateAll(const SearchState2& sstate)
{
  const NumEffectState& state = sstate.state();
  MoveLogProbVector all;
  feature_set().generateLogProb(state, env, limit, all);
  for (size_t i=0; i<all.size(); ++i) {
    const Move m = all[i].getMove();
    if (ShouldPromoteCut::canIgnoreAndNotDrop<P>(m))
      continue;
    int limit = all[i].getLogProb();
    if (all[i].getMove().capturePtype() != PTYPE_EMPTY
	|| all[i].getMove().isPromote())
      limit = std::min(limit, 400);
    if (limit <= this->limit
	&& marker.registerIfNew(state, m)) {
#ifdef SAFE_MOVE_ONLY
      if (! m.isDrop()) {
	if (env.my_pin.test(state.getPieceOnBoard(m.from()).number())
	    && move_classifier::KingOpenMove<P>::isMember(state, m.ptype(), m.from(), m.to()))
	  continue;
	if (m.ptype() == KING && state.hasEffectBy(alt(P), m.to()))
	  continue;
      }
#endif
      moves.push_back(MoveLogProb(all[i].getMove(), limit));
    }
  }
}

void osl::search::
MoveGenerator::generateAll(Player P, const SearchState2& state, 
			   category::analyzer::CategoryMoveVector& out)
{
  assert(moves.size() == 0);

  for (int i=0; i<FINISH; ++i) {
    if (! Generators[playerToIndex(P)][i])
      continue;
    (this->*Generators[playerToIndex(P)][i])(state);
    out.push_front(category::analyzer::CategoryMoves(moves, GeneratorNames[i]));
    bool generated = moves.size();
    moves.clear();    
    if (i == KING_ESCAPE && generated)
      break;
  }
  out.reverse();
}

template <osl::Player P>
void osl::search::
MoveGenerator::generateAll(const SearchState2& state, MoveLogProbVector& out) 
{
  for (MoveLogProb m = nextTacticalMove<P>(state);
       m.validMove(); m = nextTacticalMove<P>(state)) {
    out.push_back(m);
  }
  for (MoveLogProb m = nextMove<P>(state); m.validMove(); 
       m = nextMove<P>(state)) {
    out.push_back(m);
  }
}

void osl::search::
MoveGenerator::generateAll(Player P, const SearchState2& state, MoveLogProbVector& out) 
{
  if (P==BLACK)
    generateAll<BLACK>(state, out);
  else
    generateAll<WHITE>(state, out);
}

namespace osl
{
  namespace search
  {
    template const MoveLogProb MoveGenerator::nextMoveWithGeneration<BLACK>(const SearchState2&);
    template const MoveLogProb MoveGenerator::nextMoveWithGeneration<WHITE>(const SearchState2&);

    template const MoveLogProb MoveGenerator::nextTacticalMoveWithGeneration<BLACK>(const SearchState2&);
    template const MoveLogProb MoveGenerator::nextTacticalMoveWithGeneration<WHITE>(const SearchState2&);
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
