#include "osl/rating/featureSet.h"
#include "osl/rating/ratingEnv.h"
#include "osl/eval/progressEval.h"
#include "osl/effect_util/sendOffPosition.h"
#include "osl/effect_util/effectUtil.h"
#include "osl/record/csaRecord.h"
#include "osl/record/csaIOError.h"
#include "osl/apply_move/applyMove.h"
#include "osl/misc/perfmon.h"
#include "osl/stat/histogram.h"
#include "osl/stat/variance.h"

#include <boost/format.hpp>
#include <string>
#include <iostream>
#include <iomanip>
#include <cmath>
using namespace osl;
using namespace osl::rating;

void usage(const char *prog)
{
  using namespace std;
  cerr << "Usage: " << prog << " [-v] [-f skip] csafiles\n"
       << endl;
  exit(1);
}

size_t first_skip = 0;
int verbose = 0;

struct KeepMin
{
  int min;
  explicit KeepMin(int initial = 10000) : min(initial)
  {
  }
  void add(int value) { min = std::min(value, min); }
  int value() const { return min; }
};

struct KeepMax
{
  int max;
  explicit KeepMax(int initial = -10000) : max(initial)
  {
  }
  void add(int value) { max = std::max(value, max); }
  int value() const { return max; }
};

stat::Average moves, probs, order, top_score, selected_score;
const int width = 4, length = 20;
stat::Histogram moves_histogram(width, length), selected_histogram(width, length);
stat::Histogram all_moves_histogram(width, length);
const int sc_width = 100, sc_length = 12, sc_start = -500;
stat::Histogram takeback_histogram(sc_width, sc_length, sc_start), selected_takeback(sc_width, sc_length, sc_start);
stat::Histogram takeback_order(1, 10), takeback_order_all(1, 10), takeback_order_selected(1, 10);
stat::Histogram seeplus_histogram(sc_width, sc_length, sc_start), selected_seeplus(sc_width, sc_length, sc_start);
stat::Histogram seeplus_order(1, 10), seeplus_order_all(1, 10), seeplus_order_selected(1, 10);
stat::Histogram king_escape_histogram(sc_width, sc_length, sc_start), selected_king_escape(sc_width, sc_length, sc_start);
stat::Histogram kingescape_order(1, 10), kingescape_order_all(1, 10), kingescape_order_selected(1, 10);
stat::Histogram score_histogram(sc_width, sc_length, sc_start), selected_score_histogram(sc_width, sc_length, sc_start);
stat::Histogram all_score_histogram(sc_width, sc_length, sc_start);
stat::Histogram rscore_histogram(sc_width, sc_length), rselected_score_histogram(sc_width, sc_length);
stat::Histogram rall_score_histogram(sc_width, sc_length);
KeepMin min_selected, min_top;
KeepMax max_notakeback, max_nocapture;
const int sc_length_2d = sc_length+2;
const int sc_start_2d = -100;

namespace osl
{
  void showLogProb(const stat::Histogram& numerator, const stat::Histogram& denominator)
  {
    assert(numerator.width() == denominator.width());
    assert(numerator.length() == denominator.length());
    assert(numerator.start() == denominator.start());
    stat::Histogram logprob(numerator.width(), numerator.length(), numerator.start());
    for (int i=0; i<numerator.length(); ++i) {
      const double n = numerator.frequency(i);
      const double d = denominator.frequency(i);
      const double prob = n / d;
      logprob.frequency(i) = d > 20 ? static_cast<int>(-100.0*log(prob)/log(2.0)) : 0;
    }  
    logprob.show(std::cerr);
  }
  void showLogProb(const stat::Histogram& numerator, const stat::Histogram& denom1, const stat::Histogram& denom2)
  {
    assert(numerator.width() == denom1.width());
    assert(numerator.length() == denom1.length());
    assert(numerator.start() == denom1.start());
    assert(denom1.width() == denom2.width() && denom1.length() == denom2.length() && denom1.start() == denom2.start());
    stat::Histogram l1(numerator.width(), numerator.length(), numerator.start()), 
      l2(numerator.width(), numerator.length(), numerator.start());
    for (int i=0; i<numerator.length(); ++i) {
      const double n = numerator.frequency(i);
      const double d1 = denom1.frequency(i);
      const double d2 = denom2.frequency(i);
      const double p1 = n / d1;
      const double p2 = n / d2;
      l1.frequency(i) = d1 > 20 ? static_cast<int>(-100.0*log(p1)/log(2.0)) : 0;
      l2.frequency(i) = d2 > 20 ? static_cast<int>(-100.0*log(p2)/log(2.0)) : 0;
    }  
    int value=l1.start();
    for (size_t i=0; i<l1.length(); ++i, value+=l1.width()) {
      std::cerr << std::setw(5) << value << " - " 
		<< std::setw(5) << value+(int)l1.width();
      std::cerr << " " << std::setw(8) << l1.frequency(i)
		<< " " << std::setw(8) << l2.frequency(i) << "\n";
    }
  }
  enum Property { 
    All, 
    /** ܤμ֤ */
    TakeBack, 
    /** 2ܤμ֤ */
    TakeBack2, 
    /** ֤ǤʤƬ */
    NoTakeBack, 
    /** ֤ʳζƬ */
    SeePlus, 
    SeePlus2, 
    /** 3ܤζޤ϶ʳƬ */
    SeePlusX, 
    /** ֤ǤǤʤƬ */
    NoSeePlus 
  };
  size_t find(Property property, const NumEffectState& state, const RatingEnv& e,
	      const RatedMoveVector& moves, Move selected)
  {
    if (moves.empty())
      return 1;
    size_t i = 0;
    if (property == TakeBack || property == TakeBack2) {
      if (! e.history.lastMove().isNormal())
	return moves.size();
      for (; i<moves.size(); ++i)
	if (moves[i].move().to() == e.history.lastMove().to())
	  break;
    } 
    if (property == TakeBack2) {
      ++i;
      for (; i<moves.size(); ++i)
	if (moves[i].move().to() == e.history.lastMove().to())
	  break;
    }
    if (property == NoTakeBack) {
      if (e.history.lastMove().isNormal()) {
	for (; i<moves.size(); ++i)
	  if (moves[i].move().to() != e.history.lastMove().to())
	    break;
      }
    }
    if (property == SeePlus || property == SeePlus2) {
      for (; i<moves.size(); ++i) {
	if (e.history.lastMove().isNormal() && moves[i].move().to() == e.history.lastMove().to())
	  continue;
	if (PieceEval::computeDiffAfterMoveForRP(state, moves[i].move()) > 0)
	  break;
      }
    }
    if (property == SeePlus2) {
      ++i;
      for (; i<moves.size(); ++i) {
	if (PieceEval::computeDiffAfterMoveForRP(state, moves[i].move()) > 0)
	  break;
      }
    }
    if (property == SeePlusX) {
      int num_seeplus=0;
      for (; i<moves.size(); ++i) {
	if (e.history.lastMove().isNormal() && moves[i].move().to() == e.history.lastMove().to())
	  continue;
	if (PieceEval::computeDiffAfterMoveForRP(state, moves[i].move()) > 0) {
	  if (++num_seeplus <= 1) // use 2 for 3rd see+
	    continue;
	}
	break;
      }
    }
    if (property == NoSeePlus) {
      for (; i<moves.size(); ++i) {
	if (e.history.lastMove().isNormal() && moves[i].move().to() == e.history.lastMove().to())
	  continue;
	if (PieceEval::computeDiffAfterMoveForRP(state, moves[i].move()) > 0)
	  continue;
	break;
      }
    }
    return i;
  }
  /** ƥǥȥåפμ꤬ؤ줿Ψ */
  struct TopProb
  {
    stat::Histogram selected;
    stat::Histogram generated, generated_all;
    Property property;
    TopProb(Property p) 
      : selected(sc_width,sc_length+1,200), generated(sc_width,sc_length+1,200), 
	generated_all(sc_width,sc_length+1,200), 
	property(p)
    {
    }
    void add(const NumEffectState& state, const RatingEnv& e,
	     const RatedMoveVector& moves, Move selected) 
    {
      const size_t i = find(property, state, e, moves, selected);
      if (i >= moves.size())
	return;
      generated_all.add(moves[i].rating());
      const RatedMove *found = moves.find(selected);
      if (found && (found - &*moves.begin()) <= i) {
	if (moves[i].move() == selected)
	  this->selected.add(moves[i].rating());
	generated.add(moves[i].rating());
      }
    }
    void show()
    {
      showLogProb(selected, generated, generated_all);
    }
  };
  /** rating Ȥζ̤ratingκͤȤκ˴Ť 2μ¸Ψ */
  struct RatingDiffRange
  {
    stat::Histogram selected;
    stat::Histogram generated;
    stat::Histogram all_generated;
    CArray<stat::Variance, sc_length_2d*sc_length_2d> variance;
    size_t first, last;
    
    RatingDiffRange(size_t f, size_t l) 
      : selected(1,sc_length_2d*sc_length_2d), generated(1,sc_length_2d*sc_length_2d), all_generated(1,sc_length_2d*sc_length_2d),
	first(f), last(l)
    {
    }

    static int index(int score, int diff) 
    {
      const int score_index = std::max(0, std::min((score - sc_start_2d) / sc_width, sc_length_2d-1));
      const int diff_index = std::min(diff / sc_width, sc_length_2d-1);
      assert(diff_index >= 0);
      return score_index*sc_length_2d + diff_index;
    }
    void add(const NumEffectState& state, const RatedMoveVector& moves, Move selected)
    {
      if (moves.empty() || EffectUtil::isKingInCheck(state.getTurn(), state))
	return;
      const int highest = moves[0].rating();
      const RatedMove *found = moves.find(selected);
      assert(found);
      if (! found)
	return;
      const size_t selected_order = found - &*moves.begin();
      if (first <= selected_order && selected_order < last) {
	const int selected_index = index(found->rating(), highest - found->rating());
	this->selected.add(selected_index);
      }
      for (size_t i=first; i<std::min(last,moves.size()); ++i) {
	const int index = this->index(moves[i].rating(), highest - moves[i].rating());
	all_generated.add(index);
	if (i <= selected_order)
	  generated.add(index);
	variance[index].add(i);
      }
    }
  
    void show(std::ostream& os)
    {
      os << "depth\n";
      for (size_t i=0; i<sc_length_2d; ++i) {
	for (size_t j=0; j<sc_length_2d; ++j) {
	  double s = selected.frequency(i*sc_length_2d+j);
	  double g = generated.frequency(i*sc_length_2d+j);
	  // os << std::setw(5) << (g > 20 ? s/g : 0.0);
	  os << std::setw(5) << (std::min(s,g) > 20 ? static_cast<int>(-100.0*log(s/g)/log(2.0)) : 0);
	}
	os << "\n";
      }
      os << "width\n";
      for (size_t i=0; i<sc_length_2d; ++i) {
	for (size_t j=0; j<sc_length_2d; ++j) {
	  double s = selected.frequency(i*sc_length_2d+j);
	  double a = all_generated.frequency(i*sc_length_2d+j);
	  // os << std::setw(5) << (a > 20 ? s/a : 0.0);
	  os << std::setw(5) << (std::min(s,a) > 20 ? static_cast<int>(-100.0*log(s/a)/log(2.0)) : 0);
	}
	os << "\n";
      }
      os << "order\n";
      for (size_t i=0; i<sc_length_2d; ++i) {
	for (size_t j=0; j<sc_length_2d; ++j) {
	  // os << std::setw(5) << std::setprecision(3) << variance[i*sc_length_2d+j].getAverage();
	  os << std::setw(5) << static_cast<int>(variance[i*sc_length_2d+j].getAverage());
	}
	os << "\n";
      }
    }
  };
  struct RatingDiff
  {
    RatingDiffRange r0, r1, r_all;
    RatingDiff() :
      r0(1,25), r1(25,800), r_all(1,800)
    {
    }
    void add(const NumEffectState& state, const RatedMoveVector& moves, Move selected)
    {
#ifdef SHOW_SPLIT_RATING
      r0.add(state, moves, selected);
      r1.add(state, moves, selected);
#endif
      r_all.add(state, moves, selected);
    }
    void show(std::ostream& os)
    {
#if SHOW_SPLIT_RATING
      r0.show(os);
      r1.show(os);
#endif
      r_all.show(os);
    }
  };
}

RatingDiff rating_diff;
TopProb top_prob(All), takeback_topprob(TakeBack), takeback2_topprob(TakeBack2), 
  no_takeback_topprob(NoTakeBack), seeplus_topprob(SeePlus), seeplus2_topprob(SeePlus2), seeplusx_topprob(SeePlusX);
CArray<stat::Variance, 8> top_rating_progress;

void test_file(const FeatureSet&, const char *filename);

int main(int argc, char **argv)
{
  const char *program_name = argv[0];
  bool error_flag = false;
  extern char *optarg;
  extern int optind;
    
  char c;
  while ((c = getopt(argc, argv, "f:vh")) != EOF)
  {
    switch(c)
    {
    case 'f':	first_skip = atoi(optarg);
      break;
    case 'v':	++verbose;
      break;
    default:	error_flag = true;
    }
  }
  argc -= optind;
  argv += optind;

  if (error_flag || (argc < 1))
    usage(program_name);

  eval::ProgressEval::setUp();
  StandardFeatureSet f;

  for (int i=0; i<argc; ++i)
  {
    if (i % 128 == 0)
      std::cerr << '.';
    test_file(f, argv[i]);
  }

  std::cout << "\n"
	    << "average moves/position " << moves.getAverage() << "\n"
	    << "average order " << order.getAverage() << "\n"
	    << "average selected score " << selected_score.getAverage() << "\n"
	    << "min selected score " << min_selected.value() << "\n"
	    << "average top score " << top_score.getAverage() << "\n"
	    << "min top score " << min_top.value() << "\n"
	    << "max top score (notakeback) " << max_notakeback.value() << "\n"
	    << "max top score (nocapture) " << max_nocapture.value() << "\n";
  if (verbose) {
    std::cerr << "order to logprob\n";
    showLogProb(selected_histogram, moves_histogram, all_moves_histogram);
    std::cerr << "score to logprob (takeback)\n";
    showLogProb(selected_takeback, takeback_histogram);
    std::cerr << "order to logprob (takeback)\n";
    showLogProb(takeback_order_selected, takeback_order, takeback_order_all);
    std::cerr << "score to logprob (see+)\n";
    showLogProb(selected_seeplus, seeplus_histogram);
    std::cerr << "order to logprob (seeplus)\n";
    showLogProb(seeplus_order_selected, seeplus_order, seeplus_order_all);
    std::cerr << "score to logprob (king_escape)\n";
    showLogProb(selected_king_escape, king_escape_histogram);
    std::cerr << "order to logprob (kingescape)\n";
    showLogProb(kingescape_order_selected, kingescape_order, kingescape_order_all);
    if (verbose > 1) {
      std::cerr << "score to logprob (all)\n";
      showLogProb(selected_score_histogram, score_histogram, all_score_histogram);
      std::cerr << "relative score to logprob (all)\n";
      showLogProb(rselected_score_histogram, rscore_histogram, rall_score_histogram);
    }
  }
  rating_diff.show(std::cerr);
  std::cerr << "top move\n";
  top_prob.show();
  std::cerr << "top move (takeback)\n";
  takeback_topprob.show();
  std::cerr << "top move (2nd takeback)\n";
  takeback2_topprob.show();
  if (verbose > 1) {
    std::cerr << "top move (no takeback)\n";
    no_takeback_topprob.show();
  }
  std::cerr << "top move (see+)\n";
  seeplus_topprob.show();
  std::cerr << "top move (2nd see+)\n";
  seeplus2_topprob.show();
  std::cerr << "top move (2nd see+ or no see+)\n";
  seeplusx_topprob.show();
  std::cerr << "top rating for each progress (/2)\n";
  for (size_t i=0; i<top_rating_progress.size(); ++i)
    std::cerr << "progress/2 " << i << "\tave. " << std::setprecision(3) << top_rating_progress[i].getAverage()
	      << "\tsigma " << sqrt(top_rating_progress[i].variance()) << "\n";
}

/* ------------------------------------------------------------------------- */

size_t num_positions = 0;
void test_position(const FeatureSet& f, Move next_move, Move last_move, const RatingEnv& env,
		   const NumEffectState& state, const eval::ProgressEval& eval)
{
  const bool in_check = EffectUtil::isKingInCheck(state.getTurn(), state);
  RatedMoveVector moves;
  f.generateRating(state, env, moves);

  rating_diff.add(state, moves, next_move);
  top_prob.add(state, env, moves, next_move);
  takeback_topprob.add(state, env, moves, next_move);
  takeback2_topprob.add(state, env, moves, next_move);
  no_takeback_topprob.add(state, env, moves, next_move);
  seeplus_topprob.add(state, env, moves, next_move);
  seeplus2_topprob.add(state, env, moves, next_move);
  seeplusx_topprob.add(state, env, moves, next_move);
  
  bool notakeback_added = in_check, nocapture_added = in_check;
  if (moves.size()) {
    const int highest = moves[0].rating();
    min_top.add(highest);
    top_score.add(highest);
    if (! in_check) {
      size_t index = find(NoSeePlus, state, env, moves, next_move);
      if (index < moves.size())
	top_rating_progress[eval.progress16().value()/2].add(moves[index].rating());
    }
    if (! notakeback_added
	&& moves[0].move().to() != last_move.to()) {
      nocapture_added = true;
      max_notakeback.add(highest);
    }
    if (! nocapture_added
	&& moves[0].move().capturePtype() == PTYPE_EMPTY
	&& ! moves[0].move().isPromote()) {
      nocapture_added = true;
      max_nocapture.add(highest);
    }
  }
  const RatedMove *p = moves.find(next_move);
  int count = moves.size();
  int order = p ? p - &*moves.begin() +1 : count;
  if (p) {
    ::order.add(order);
    selected_histogram.add(order);
    selected_score.add(p->rating());
    min_selected.add(p->rating());
    if (p->rating() < -2000) {
      std::cerr << state << "selected " << *p << "\n" << moves;
    }
  }
  else {
    ::order.add(count);
  }
  for (int i=0; i<order; ++i)
    moves_histogram.add(i);
  for (size_t i=0; i<moves.size(); ++i)
    all_moves_histogram.add(i);
  ::moves.add(count);
  ++num_positions;

  if (in_check) {
    for (int i=0; i<count; ++i) {
      if (i < order) {
	king_escape_histogram.add(moves[i].rating());
	kingescape_order.add(i);
	if (moves[i].move() == next_move) {
	  selected_king_escape.add(moves[i].rating());
	  kingescape_order_selected.add(i);
	}
      }
      kingescape_order_all.add(i);
    }
  }
  else {
    int j=0;
    for (int i=0; i<count; ++i) {
      if (moves[i].move().to() != last_move.to())
	continue;
      if (i < order) {
	takeback_histogram.add(moves[i].rating());
	takeback_order.add(j);
	if (moves[i].move() == next_move) {
	  selected_takeback.add(moves[i].rating());
	  takeback_order_selected.add(j);
	}
      }
      takeback_order_all.add(j);
      ++j;
    }
    j=0;
    for (int i=0; i<count; ++i) {
      if (moves[i].move().to() == last_move.to())
	continue;
      if (! (moves[i].move().capturePtype() != PTYPE_EMPTY
	     || moves[i].move().isPromote())
	  || PieceEval::computeDiffAfterMoveForRP(state, moves[i].move()) <= 0)
	continue;
      if (i<order) {
	seeplus_histogram.add(moves[i].rating());
	seeplus_order.add(j);
	if (moves[i].move() == next_move) {
	  selected_seeplus.add(moves[i].rating());
	  seeplus_order_selected.add(j);
	}
      }
      seeplus_order_all.add(j);
      ++j;
    }

    for (int i=0; i<order; ++i) {
      score_histogram.add(moves[i].rating());
      if (moves[i].move() == next_move)
	selected_score_histogram.add(moves[i].rating());
    }
    for (size_t i=0; i<moves.size(); ++i) {
      all_score_histogram.add(moves[i].rating());
      if (! notakeback_added && moves[i].move().to() != last_move.to()) {
	notakeback_added = true;
	max_notakeback.add(moves[i].rating());
      }
      if (! nocapture_added && moves[i].move().capturePtype() == PTYPE_EMPTY
	  && ! moves[i].move().isPromote()) {
	nocapture_added = true;
	max_nocapture.add(moves[i].rating());
      }
    }
    if (moves[0].move() != next_move) {
      const int top_score = moves[0].rating();
      for (int i=1; i<order; ++i) {
	rscore_histogram.add(top_score - moves[i].rating());
	if (moves[i].move() == next_move)
	  rselected_score_histogram.add(top_score - moves[i].rating());
      }
      for (size_t i=1; i<moves.size(); ++i) {
	rall_score_histogram.add(top_score - moves[i].rating());
      }
    }
  }
}

void test_file(const FeatureSet& f, const char *filename)
{
  Record rec;
  try {
    rec = CsaFile(filename).getRecord();
  }
  catch (CsaIOError& e) {
    std::cerr << "skip " << filename <<"\n";
    std::cerr << e.what() << "\n";
    return;
  }
  catch (...) {
    throw;
  }
  
  NumEffectState state(rec.getInitialState());
  const osl::stl::vector<osl::Move> moves=rec.getMoves();

  RatingEnv env;
  eval::ProgressEval eval(state);
  for (size_t i=0; i<moves.size(); ++i) {
    const Move move = moves[i];
    assert(state.isValidMove(move));
    if (i >= first_skip) {
      test_position(f, moves[i], (i>0 ? moves[i-1] : Move::PASS(alt(moves[i].player()))),
		    env, state, eval);
    }
    ApplyMoveOfTurn::doMove(state, move);
    eval.update(state, move);
    env.history.push(move);
    env.sendoffs.clear();
    const Position king_position = state.getKingPosition(alt(state.getTurn()));
    effect_util::SendOffPosition::find(state.getTurn(), state, king_position, 	
				       env.sendoffs);
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
